from thunno2.commands import commands, DIGRAPHS, get_a_function, Void
from thunno2.constants import CONSTANTS
from thunno2.codepage import codepage_index


"""Splits Thunno 2 code into tokens to make it easier for the interpreter"""


def tokenise(code, expected_end=''):
    index = 0
    ret = []
    while index < len(code):
        char = code[index]
        if char in commands:
            ret.append((char, 'command', commands[char]))
        elif char in DIGRAPHS:
            index += 1
            try:
                x = code[index]
                ret.append((char, 'digraph', get_a_function(char + x)))
            except:
                pass
        elif char in '0123456789.':
            s = char
            index += 1
            try:
                while code[index] in '0123456789.':
                    s += code[index]
                    index += 1
            except:
                pass
            index -= 1
            try:
                while s[0] == '0':
                    ret.append(('0', 'number', 0))
                    try:
                        s = s[1:]
                    except:
                        break
                if s == '.':
                    ret.append(('.', 'number', 0.5))
                else:
                    x = eval(s)
                    if s.endswith('.'):
                        x += 0.5
                    ret.append((s, 'number', x))
            except:
                pass
        elif char == '"':
            s = char
            index += 1
            try:
                while (code[index] != '"') or (code[index - 1] == '\\'):
                    s += code[index]
                    index += 1
                s += code[index]
            except:
                s += '"'
            try:
                ret.append((s, 'string', eval(s).replace('¶', '\n')))
            except:
                ret.append((s, 'string', s[1:-1].replace('¶', '\n')))
        elif char == '\'':
            index += 1
            try:
                x = code[index]
                ret.append(('\'' + x, 'one character', x))
            except:
                ret.append(('\'', 'one character', ''))
        elif char == '`':
            index += 2
            x = code[index - 1: index + 1]
            ret.append(('`' + x, 'two characters', x))
        elif char == 'ʋ':
            index += 3
            x = code[index - 2: index + 1]
            ret.append(('ʋ' + x, 'three characters', x))
        elif char == '[':
            s = char
            index += 1
            try:
                in_string = ''
                nests = 1
                while nests:
                    c = code[index]
                    s += c
                    if in_string == '' and c in ('"', "'"):
                        in_string = c
                    elif in_string != '' and c == in_string:
                        if code[index - 1] != '\\':
                            in_string = ''
                    if in_string == '':
                        if c == '[':
                            nests += 1
                        elif c == ']':
                            nests -= 1
                    index += 1
            except:
                s += ']'
            try:
                ret.append((s, 'list', eval(s)))
            except:
                ret.append((s, 'list', s))
        elif char == ']':
            ret.append((']', 'list', []))
        elif char == '#':
            index += 1
            try:
                if code[index] == ' ':
                    while code[index] not in '¶\n':
                        index += 1
                elif code[index] == '{':
                    while code[index] != '#' or code[index - 1] != '}':
                        index += 1
            except:
                pass
        elif char == '“':
            compressed_string = ''
            index += 1
            try:
                while code[index] != char:
                    compressed_string += code[index]
                    index += 1
            except:
                pass
            ret.append((char + compressed_string + char, 'lowercase alphabetic compression', compressed_string))
        elif char == '”':
            compressed_string = ''
            index += 1
            try:
                while code[index] != char:
                    compressed_string += code[index]
                    index += 1
            except:
                pass
            ret.append((char + compressed_string + char, 'title case alphabetic compression', compressed_string))
        elif char == '‘':
            compressed_string = ''
            index += 1
            try:
                while code[index] != char:
                    compressed_string += code[index]
                    index += 1
            except:
                pass
            ret.append((char + compressed_string + char, 'lowercase dictionary compression', compressed_string))
        elif char == '’':
            compressed_string = ''
            index += 1
            try:
                while code[index] != char:
                    compressed_string += code[index]
                    index += 1
            except:
                pass
            ret.append((char + compressed_string + char, 'title case dictionary compression', compressed_string))
        elif char == '»':
            compressed_string = ''
            index += 1
            try:
                while code[index] != char:
                    compressed_string += code[index]
                    index += 1
            except:
                pass
            ret.append((char + compressed_string + char, 'compressed number', compressed_string))
        elif char == '«':
            compressed_string = code[index + 1:index + 3]
            index += 2
            ret.append((char + compressed_string, 'small compressed number', compressed_string))
        elif char == '¿':
            compressed_string = ''
            index += 1
            try:
                while code[index] != char:
                    compressed_string += code[index]
                    index += 1
            except:
                pass
            ret.append((char + compressed_string + char, 'compressed list', compressed_string))
        elif char == '¡':
            index += 1
            try:
                var = code[index]
            except:
                var = ''
            ret.append((char + var, 'variable get', var))
        elif char == '!':
            index += 1
            try:
                var = code[index]
            except:
                var = ''
            ret.append((char + var, 'variable set', var))
        elif char == '€':
            index += 1
            cmd = code[index]
            if cmd in DIGRAPHS:
                index += 1
                cmd += code[index]
            func = get_a_function(cmd)
            ret.append((char + cmd, 'single function map', func))
        elif char == 'ȷ':
            index += 1
            cmd = code[index]
            if cmd in DIGRAPHS:
                index += 1
                cmd += code[index]
            func = get_a_function(cmd)
            ret.append((char + cmd, 'outer product', func))
        elif char == 'œ':
            index += 1
            cmd = code[index]
            if cmd in DIGRAPHS:
                index += 1
                cmd += code[index]
            func = get_a_function(cmd)
            ret.append((char + cmd, 'single function filter', func))
        elif char == 'þ':
            index += 1
            cmd = code[index]
            if cmd in DIGRAPHS:
                index += 1
                cmd += code[index]
            func = get_a_function(cmd)
            ret.append((char + cmd, 'single function sort by', func))
        elif char == 'ñ':
            index += 1
            cmd = code[index]
            if cmd in DIGRAPHS:
                index += 1
                cmd += code[index]
            func = get_a_function(cmd)
            ret.append((char + cmd, 'single function group by', func))
        elif char == 'n':
            ret.append((char, 'context variable', 0))
        elif char == 'ṅ':
            ret.append((char, 'iteration index', 0))
        elif char == 'x':
            ret.append((char, 'get x', 0))
        elif char == 'y':
            ret.append((char, 'get y', 0))
        elif char == 'X':
            ret.append((char, 'set x', 0))
        elif char == 'Y':
            ret.append((char, 'set y', 0))
        elif char == 'ẋ':
            ret.append((char, 'set x without popping', 0))
        elif char == 'ẏ':
            ret.append((char, 'set y without popping', 0))
        elif char == 'ẋ':
            ret.append((char, 'increment x', 0))
        elif char == 'ẏ':
            ret.append((char, 'increment y', 0))
        elif char == 'Ȥ':
            ret.append((char, 'get global array', 0))
        elif char == 'ȥ':
            ret.append((char, 'add to global array', 0))
        elif char == 'K':
            ret.append((char, 'stack', 0))
        elif char == 'k':
            index += 1
            try:
                if code[index] in CONSTANTS:
                    x = code[index]
                    ret.append((char + x, 'constant', CONSTANTS[x]))
            except:
                pass
        elif char == 'ṇ':
            index += 1
            try:
                x = code[index]
                ret.append((char + x, 'codepage compression', next(codepage_index(x)) + 101))
            except:
                pass
        elif char == 'q':
            ret.append((char, 'quit', 0))
        elif char == '$':
            ret.append((char, 'next input', 0))
        elif char == '¤':
            ret.append((char, 'input list', 0))
        elif char == '°':
            ret.append((char, 'first input', 0))
        elif char == '¹':
            ret.append((char, 'second input', 0))
        elif char == '⁶':
            ret.append((char, 'third input', 0))
        elif char == '⁷':
            ret.append((char, 'third last input', 0))
        elif char == '⁸':
            ret.append((char, 'second last input', 0))
        elif char == '⁹':
            ret.append((char, 'last input', 0))
        elif char == '£':
            ret.append((char, 'print', 0))
        elif char == '¢':
            ret.append((char, 'print without newline', 0))
        elif char == 'ß':
            ret.append((char, 'print without popping', 0))
        elif char == 'ı':
            i, r = tokenise(code[index + 1:], expected_end=';')
            index += i
            ret.append((char, 'map', r))
        elif char == 'æ':
            i, r = tokenise(code[index + 1:], expected_end=';')
            index += i
            ret.append((char, 'filter', r))
        elif char == 'Þ':
            i, r = tokenise(code[index + 1:], expected_end=';')
            index += i
            ret.append((char, 'sort by', r))
        elif char == 'Ñ':
            i, r = tokenise(code[index + 1:], expected_end=';')
            index += i
            ret.append((char, 'group by', r))
        elif char == '¥':
            i, r = tokenise(code[index + 1:], expected_end=';')
            index += i
            ret.append((char, 'fixed point', r))
        elif char == 'Ƙ':
            i, r = tokenise(code[index + 1:], expected_end=';')
            index += i
            ret.append((char, 'first n integers', r))
        elif char == 'Ʋ':
            i, r = tokenise(code[index + 1:], expected_end=';')
            index += i
            ret.append((char, 'cumulative reduce by', r))
        elif char == '{':
            i, r = tokenise(code[index + 1:], expected_end='}')
            index += i
            ret.append((char, 'for loop', r))
        elif char == '(':
            i, r1 = tokenise(code[index + 1:], expected_end=';)')
            index += i + 1
            r2 = []
            try:
                if code[index] == ';':
                    i, r2 = tokenise(code[index + 1:], expected_end=')')
                    index += i
            except:
                pass
            ret.append((char, 'while loop', (r1, r2)))
        elif char == '?':
            i, r1 = tokenise(code[index + 1:], expected_end=':;')
            index += i + 1
            r2 = []
            try:
                if code[index] == ':':
                    i, r2 = tokenise(code[index + 1:], expected_end=';')
                    index += i
            except:
                pass
            ret.append((char, 'if statement', (r1, r2)))
        elif char == 'Ɓ':
            index += 1
            cmd = code[index]
            if cmd in DIGRAPHS:
                index += 1
                cmd += code[index]
            func = get_a_function(cmd)
            ret.append((char + cmd, 'execute without popping', func))
        elif char == 'ç':
            try:
                index += 1
                cmd1 = code[index]
                if cmd1 in DIGRAPHS:
                    index += 1
                    cmd1 += code[index]
                func1 = get_a_function(cmd1)
            except:
                func1, cmd1 = Void, ''
            try:
                index += 1
                cmd2 = code[index]
                if cmd2 in DIGRAPHS:
                    index += 1
                    cmd2 += code[index]
                func2 = get_a_function(cmd2)
            except:
                func2, cmd2 = Void, ''
            ret.append((char + cmd1 + cmd2, 'pair apply', (func1, func2)))
        elif char == 'Ç':
            try:
                index += 1
                cmd1 = code[index]
                if cmd1 in DIGRAPHS:
                    index += 1
                    cmd1 += code[index]
                func1 = get_a_function(cmd1)
            except:
                func1, cmd1 = Void, ''
            try:
                index += 1
                cmd2 = code[index]
                if cmd2 in DIGRAPHS:
                    index += 1
                    cmd2 += code[index]
                func2 = get_a_function(cmd2)
            except:
                func2, cmd2 = Void, ''
            ret.append((char + cmd1 + cmd2, 'two function map', (func1, func2)))
        elif char in expected_end:
            return index, ret
        index += 1
    return index + 1, ret



"""
NOTE: This is just an experiment.

Each command in Thunno 2 has been assigned at least one token.
Using those tokens we can retrieve the original command.
"""

from thunno2.interpreter import commands
import sys


def get_command(token):
    for cmd, ovld in commands.items():
        if token.lower() in ovld.keywords:
            return cmd
    print(f"Couldn't find a command for token {token!r}", file=sys.stderr)
    sys.exit(0)


def transpile(code):
    r = ''
    for word in code.split():
        if word[:1] == '\\':
            r += word[1:]
        else:
            r += get_command(word)
    return r


def test():
    l = [token for cmd, ovld in commands.items() for token in ovld.keywords]
    for t in l:
        assert len(t) >= 2, f"Token {t!r} too short"
        assert l.count(t) == 1, f"Token {t!r} not unique"
        assert t == t.lower(), f"Token {t!r} not lowercase"
        assert t.replace('_', '').isalnum(), f"Token {t!r} invalid"
    print('[TOKENS]: Passed')


test()



from thunno2.helpers import *
import sys


class TerminateProgramException(BaseException):
    pass


class Stack(list):
    # Push one item to the front
    def push(self, item):
        self.insert(0, item)

    # Remove n items from the front
    def rmv(self, num, default=0):
        for _ in range(num):
            if list(self):
                yield self.pop(0)
            else:
                if ctx.other_il:
                    x = ctx.other_il.pop(0)
                    ctx.other_il.append(x)
                    yield x
                else:
                    yield default
    
    
class Context:
    def __init__(self):
        self.stack = Stack()
        self.og_input_list = []
        self.other_il = []
        self.implicit_print = True
        self.warnings = False
        

ctx = Context()


class VoidType:

    def __call__(self, *args, **kwds):
        return ()

    def __eq__(self, other):
        return isinstance(other, VoidType)

    def __ne__(self, other):
        return not isinstance(other, VoidType)

    def __bool__(self):
        return False

    def __int__(self):
        return 0

    def __float__(self):
        return 0.0

    def __str__(self):
        return 'Void'

    def __repr__(self):
        return 'Void'


Void = VoidType()


def recursively_vectorise(lst, dct):
    for item in lst:
        if isinstance(item, list):
            yield list(recursively_vectorise(item, dct))
        else:
            for key, val in dct.items():
                if isinstance(item, key):
                    try:
                        yield val(item)
                    except Exception as E:
                        if ctx.warnings:
                            print('TRACEBACK: [ERROR - CAUGHT]', file=sys.stderr)
                            print(E, file=sys.stderr)
                        pass
                    break
            else:
                if ctx.warnings:
                    print('TRACEBACK: [UNEXPECTED TYPE]', file=sys.stderr)
                    print('Got', type(item), file=sys.stderr)
                yield Void


def recursively_distribute(x, y, dct, main=1):
    if (not isinstance(x, list)) and (not isinstance(y, list)):
        for (a, b), f in dct.items():
            if isinstance(x, a) and isinstance(y, b):
                yield f(x, y)
                break
        else:
            if ctx.warnings:
                print('TRACEBACK: [UNEXPECTED TYPE]', file=sys.stderr)
                print('Got', (type(x), type(y)), file=sys.stderr)
            yield Void
    elif isinstance(x, list) and not isinstance(y, list):
        r = []
        for i in x:
            if isinstance(i, list):
                r += next(recursively_distribute(i, y, dct, 0)),
            else:
                for (a, b), f in dct.items():
                    if isinstance(i, a) and isinstance(y, b):
                        r += f(i, y),
                        break
                else:
                    if ctx.warnings:
                        print('TRACEBACK: [UNEXPECTED TYPE]', file=sys.stderr)
                        print('Got', (type(i), type(y)), file=sys.stderr)
        yield r
    elif isinstance(y, list) and not isinstance(x, list):
        r = []
        for i in y:
            if isinstance(i, list):
                r += next(recursively_distribute(x, i, dct, 0)),
            else:
                for (a, b), f in dct.items():
                    if isinstance(x, a) and isinstance(i, b):
                        r += f(x, i),
                        break
                else:
                    if ctx.warnings:
                        print('TRACEBACK: [UNEXPECTED TYPE]', file=sys.stderr)
                        print('Got', (type(x), type(i)), file=sys.stderr)
        yield r
    else:
        if not x:
            yield from y
        elif not y:
            yield from x
        else:
            if main:
                yield list(recursively_distribute(x[0], y[0], dct, 0)) + list(
                    recursively_distribute(x[1:], y[1:], dct, 0))
            else:
                yield from list(recursively_distribute(x[0], y[0], dct, 0)) + list(
                    recursively_distribute(x[1:], y[1:], dct, 0))


def repeat_to_longest(*lsts):
    max_len = max(map(len, lsts))
    for inner_list in lsts:
        if not inner_list:
            yield Void
        else:
            yield (inner_list * max_len)[:max_len]


class Overload:
    def __init__(self,
                 pop,  #: int,  # How many items do we need to pop?
                 funcs,  #: dict[tuple[type], Callable],  # A mapping of the type overloads to functions
                 vectorise,  #: #int,    # What vectorisation type?
                 #  0 = no vectorisation,
                 #  1 = monadic recursive vectorisation,
                 #  2 = dyadic recursive vectorisation,
                 #  3 = triadic one-level vectorisation
                 keywords  #: tuple[str]    # This is just an experiment with tokens, might not be implemented
                 ):
        self.pop = pop
        self.funcs = {k if isinstance(k, tuple) else (k,): v for k, v in funcs.items()}
        self.vectorise = vectorise
        self.keywords = keywords

    def __call__(self, pop=True):
        if pop:
            items = list(ctx.stack.rmv(self.pop))
        else:
            items = list(
                (
                        ctx.stack
                        + ctx.other_il
                        + [0] * self.pop
                )[:self.pop]
            )
        if self.vectorise == 0:
            for typ, fun in self.funcs.items():
                if all(isinstance(x, y) for x, y in zip(items, typ)):
                    try:
                        r = fun(*items)
                        if not isinstance(r, tuple):
                            return r,
                        else:
                            return r
                    except Exception as E:
                        if ctx.warnings:
                            print('TRACEBACK: [ERROR - CAUGHT]', file=sys.stderr)
                            print(E, file=sys.stderr)
                        return Void
            if ctx.warnings:
                print('TRACEBACK: [UNEXPECTED TYPE]', file=sys.stderr)
                print('Got', (*map(type, items),), file=sys.stderr)
            return Void
        elif self.vectorise == 1:
            if len(items) == 1:
                if isinstance(items[0], list):
                    try:
                        return list(recursively_vectorise(items[0], self.funcs)),
                    except Exception as E:
                        if ctx.warnings:
                            print('TRACEBACK: [ERROR - CAUGHT]', file=sys.stderr)
                            print(E, file=sys.stderr)
                        return Void
                else:
                    for typ, fun in self.funcs.items():
                        if isinstance(items[0], typ):
                            try:
                                r = fun(*items)
                                if not isinstance(r, tuple):
                                    return r,
                                else:
                                    return r
                            except Exception as E:
                                if ctx.warnings:
                                    print('TRACEBACK: [ERROR - CAUGHT]', file=sys.stderr)
                                    print(E, file=sys.stderr)
                    else:
                        if ctx.warnings:
                            print('TRACEBACK: [UNEXPECTED TYPE]', file=sys.stderr)
                            print('Got', type(items[0]), file=sys.stderr)
            else:
                if ctx.warnings:
                    print('TRACEBACK: [IMPLEMENTATION FAULT]', file=sys.stderr)
                    print('Expected 1 item, got', len(items), file=sys.stderr)
                return Void
        elif self.vectorise == 2:
            try:
                r = tuple(recursively_distribute(items[0], items[1], self.funcs))
                return r
            except Exception as E:
                if ctx.warnings:
                    print('TRACEBACK: [ERROR - CAUGHT]', file=sys.stderr)
                    print(E, file=sys.stderr)
                return Void
        elif self.vectorise == 3:
            if not any(isinstance(it, list) for it in items):
                for typ, fun in self.funcs.items():
                    if all(isinstance(x, y) for x, y in zip(items, typ)):
                        try:
                            r = fun(*items)
                            if not isinstance(r, tuple):
                                return r,
                            else:
                                return r
                        except Exception as E:
                            if ctx.warnings:
                                print('TRACEBACK: [ERROR - CAUGHT]', file=sys.stderr)
                                print(E, file=sys.stderr)
                            return Void
                else:
                    if ctx.warnings:
                        print('TRACEBACK: [UNEXPECTED TYPE]', file=sys.stderr)
                        print('Got', (*map(type, items),), file=sys.stderr)
                return Void
            lst = [i if isinstance(i, list) else [i] * safe_max_len(*items) for i in items]
            ret = []
            for il in itertools.zip_longest(*lst):
                if None in il:
                    ret.append(first_non_none(il))
                else:
                    for typ, fun in self.funcs.items():
                        if all(isinstance(x, y) for x, y in zip(il, typ)):
                            try:
                                ret.append(fun(*il))
                            except Exception as E:
                                if ctx.warnings:
                                    print('TRACEBACK: [ERROR - CAUGHT]', file=sys.stderr)
                                    print(E, file=sys.stderr)
                                pass
                            break
                    else:
                        if ctx.warnings:
                            print('TRACEBACK: [UNEXPECTED TYPE]', file=sys.stderr)
                            print('Got', (*map(type, items),), file=sys.stderr)
            return ret,
        if ctx.warnings:
            print('TRACEBACK: [IMPLEMENTATION FAULT]', file=sys.stderr)
            print('Unknown vectorisation argument', self.vectorise, file=sys.stderr)
        return Void


Number = ((int, float),)
Iterable = ((list, str),)
Any = ((int, float, list, str),)

# Which characters start digraphs?
DIGRAPHS = 'ØøÆµ'

'''DICTIONARY'''

commands = {

    'A': Overload(1, {
        Number: abs,
        str: isalpha
    }, 1, ('abs', 'absolute', 'isalpha')),

    'B': Overload(2, {
        (int, int): ntb,
        (str, int): to_custom_base_string,
        (int, str): length_to_base,
        (str, str): length_custom_base_string,
        (float, Any): pass_
    }, 2, ('to_base', 'from_decimal')),

    'C': Overload(1, {
        Number: chr2,
        str: ord2
    }, 1, ('chr', 'character', 'ord', 'ordinal')),

    'D': Overload(1, {
        Any: duplicate
    }, 0, ('dup', 'duplicate')),

    'E': Overload(1, {
        Number: is_even,
        str: eval2
    }, 1, ('even', 'is_even', 'eval')),

    'F': Overload(1, {
        Number: factors,
        str: substrings
    }, 1, ('factors', 'substrings')),

    'G': Overload(1, {
        Number: max2,
        Iterable: max3
    }, 0, ('max', 'maximum', 'greatest')),

    'H': Overload(1, {
        (int, str): from_hex,
        float: pass_
    }, 1, ('from_hex',)),

    'I': Overload(2, {
        (Number[0], Number[0]): binary_range,
        (Number[0], Iterable[0]): slice1,
        (Iterable[0], Number[0]): slice2,
        (list, list): interleave_lst,
        (str, str): interleave_str
    }, 0, ('binary_range', 'slice', 'interleave')),

    'J': Overload(1, {
        Number: str,
        Iterable: empty_join
    }, 0, ('empty_join', 'str')),

    # K is defined in the run function

    'L': Overload(1, {
        Number: lowered_range,
        str: str.lower
    }, 1, ('lowered_range', 'zero_range', 'lower', 'lowercase')),

    'M': Overload(1, {
        Number: min2,
        Iterable: min3
    }, 0, ('min', 'minimum')),

    'N': Overload(1, {
        (Number[0], str): safe_int
    }, 1, ('int', 'integer')),

    'O': Overload(1, {
        Number: two_power,
        str: str.split
    }, 1, ('two_power', 'space_split')),

    'P': Overload(1, {
        Number: is_prime,
        str: str.swapcase
    }, 1, ('prime', 'is_prime', 'swapcase')),

    'Q': Overload(2, {
        ((Number[0], str), (Number[0], str)): not_equal
    }, 2, ('not_equal', 'inequality')),

    'R': Overload(1, {
        Number: one_range,
        str: str.upper
    }, 1, ('one_range', 'range', 'upper', 'uppercase')),

    'S': Overload(1, {
        Number: digit_sum,
        Iterable: it_sum
    }, 0, ('sum',)),

    'T': Overload(0, {
        Any: (lambda: 10)
    }, 0, ('ten',)),

    'U': Overload(1, {
        Number: uniquify_num,
        str: uniquify_str,
        list: uniquify_lst
    }, 0, ('uniquify',)),

    'V': Overload(1, {
        Number: indices_where_truthy_num,
        str: rot_13,
        list: indices_where_truthy
    }, 0, ('where_truthy', 'rot13')),

    'W': Overload(1, {
        Any: wrap
    }, 0, ('wrap',)),

    # X is defined in the run function

    # Y is defined in the run function

    'Z': Overload(2, {
        (Number[0], Number[0]): range_zip1,
        (Number[0], Iterable[0]): range_zip2,
        (Iterable[0], Number[0]): range_zip3,
        (Iterable[0], Iterable[0]): zip2
    }, 0, ('zip',)),

    'a': Overload(2, {
        (Any[0], list): append,
        (list, Any[0]): swapped_append,
        (Any[0], Any[0]): append2
    }, 0, ('append',)),

    'b': Overload(2, {
        (int, (int, str)): convert_from_base,
        (str, (int, str)): convert_from_custom_base,
        (float, Any): pass_,
        (Any, float): pass_
    }, 2, ('from_base', 'to_decimal')),

    'c': Overload(2, {
        (Number[0], Number[0]): nCr,
        (Any, str): string_count,
        (Any, list): list_count
    }, 0, ('ncr', 'count')),

    'd': Overload(1, {
        Number: digits,
        str: list
    }, 1, ('digits', 'characters', 'chars')),

    'e': Overload(1, {
        Number: ten_power,
        str: comma_split
    }, 1, ('ten_power', 'comma_split')),

    'f': Overload(1, {
        Number: prime_factors,
        str: case
    }, 1, ('prime_factors', 'case')),

    'g': Overload(1, {
        Number: digits_gcd,
        str: ords_gcd,
        list: gcd
    }, 0, ('gcd', 'greatest_common_divisor')),

    'h': Overload(1, {
        Number: num_head,
        Iterable: head
    }, 0, ('head', 'first')),

    'i': Overload(2, {
        (Number[0], Number[0]): num_ind0,
        (Number[0], Iterable[0]): indexing_0,
        (Iterable[0], Number[0]): swapped_ind0,
        (str, Iterable[0]): length_ind0,
        (list, Iterable[0]): vectorised_ind0
    }, 0, ('indexing', 'zero_indexing')),

    'j': Overload(2, {
        (Any, Number[0]): num_join,
        (Any, Iterable[0]): join
    }, 0, ('join',)),

    # k is defined in the run function

    'l': Overload(1, {
        Number: num_length,
        Iterable: len
    }, 0, ('len', 'length')),

    'm': Overload(1, {
        Number: num_mean,
        str: ord_mean,
        list: mean
    }, 0, ('mean', 'avg', 'average')),

    # n is defined in the run function

    'o': Overload(2, {
        (Number[0], Number[0]): num_rmv,
        (Any[0], list): list_rmv,
        (list, Any[0]): swapped_list_rmv,
        (Any[0], str): str_rmv
    }, 0, ('rmv', 'remove')),

    'p': Overload(1, {
        Number: digit_product,
        str: ord_product,
        list: product
    }, 0, ('product', 'prod')),

    # q is defined in the run function

    'r': Overload(1, {
        Number: digit_reverse,
        Iterable: reverse
    }, 0, ('reverse',)),

    's': Overload(2, {
        (Any[0], Any[0]): swap
    }, 0, ('swap',)),

    't': Overload(1, {
        Number: num_tail,
        Iterable: tail
    }, 0, ('tail', 'last')),

    'u': Overload(0, {
        Any: (lambda: -1)
    }, 0, ('negative_one',)),

    'v': Overload(3, {
        (Any[0], Any[0], list): list_replace,
        (Any[0], Any[0], Any[0]): string_replace
    }, 0, ('replace',)),

    'w': Overload(1, {
        Number: round,
        str: remove_whitespace
    }, 1, ('round', 'remove_whitespace')),

    # x is defined in the run function

    # y is defined in the run function

    'z': Overload(1, {
        Number: num_uninterleave,
        Iterable: uninterleave
    }, 0, ('uninterleave',)),

    'Ȧ': Overload(1, {
        Number: bool2,
        str: isalphanum,
        list: any2
    }, 0, ('any', 'isalphanum')),

    'Ḃ': Overload(1, {
        (int, str): from_binary,
        float: pass_
    }, 1, ('from_binary',)),

    'Ċ': Overload(1, {
        Number: codepage_chr,
        str: codepage_ord
    }, 1, ('codepage_chr', 'codepage_character', 'codepage_ord', 'codepage_ordinal')),

    'Ḋ': Overload(2, {
        (Number[0], Number[0]): is_divisible,
        (str, Number[0]): length_divisible1,
        (Number[0], str): length_divisible2,
        (str, str): length_divisible3
    }, 2, ('divisible', 'is_divisible')),

    'Ė': Overload(1, {
        Number: inclusive_zero_range,
        Iterable: zero_enumerate
    }, 0, ('enumerate', 'inclusive_zero_range')),

    'Ḟ': Overload(1, {
        Any: recursive_flatten
    }, 0, ('flatten', 'recursive_flatten')),

    'Ġ': Overload(2, {
        (Number[0], Number[0]): dyadic_gcd,
        (str, Number[0]): ordinal_gcd1,
        (Number[0], str): ordinal_gcd2,
        (str, str): longest_common_substring
    }, 2, ('dyadic_gcd',)),

    'Ḣ': Overload(1, {
        Number: num_head_extract,
        Iterable: head_extract
    }, 0, ('head_extract',)),

    'İ': Overload(2, {
        (Number[0], Number[0]): num_ind1,
        (Number[0], Iterable[0]): indexing_1,
        (Iterable[0], Number[0]): swapped_ind1,
        (str, Iterable[0]): length_ind1,
        (list, Iterable[0]): vectorised_ind1
    }, 0, ('one_indexing',)),

    'Ŀ': Overload(3, {
        (Any[0], Number[0], Any[0]): ljust1,
        (Number[0], Any[0], Any[0]): ljust2,
        (Any[0], Any[0], Any[0]): ljust3
    }, 3, ('ljust', 'left_justify')),

    'Ṁ': Overload(1, {
        Number: num_mode,
        Iterable: mode
    }, 0, ('mode', 'most_common_element')),

    'Ṅ': Overload(1, {
        Number: negate,
        str: rle
    }, 1, ('negate', 'neg', 'rle', 'run_length_encode')),

    'Ȯ': Overload(2, {
        ((str, int, float), Iterable[0]): index_of,
        (list, (str, int, float)): swapped_index_of,
        (Number[0], (str, int, float)): num_index_of,
        (list, Iterable[0]): vectorised_index_of
    }, 0, ('index_of', 'find')),

    'Ṗ': Overload(2, {
        (Iterable[0], Iterable[0]): cartesian_product,
        (Number[0], Iterable[0]): range_product1,
        (Iterable[0], Number[0]): range_product2,
        (Number[0], Number[0]): range_product3
    }, 0, ('cartesian_product',)),

    'Ṙ': Overload(1, {
        Number: str,
        str: string_repr,
        list: list_repr
    }, 0, ('repr',)),

    'Ṡ': Overload(1, {
        Number: digits_sort,
        str: str_sort,
        list: sort
    }, 0, ('sort',)),

    'Ṫ': Overload(1, {
        Number: num_tail_extract,
        Iterable: tail_extract
    }, 0, ('tail_extract',)),

    'Ẇ': Overload(2, {
        (Iterable[0], Number[0]): split1,
        (Number[0], Iterable[0]): split2,
        (Number[0], Number[0]): chunk1,
        (Iterable[0], Iterable[0]): chunk2
    }, 0, ('chunk', 'chunk_split')),

    # Ẋ is defined in the run function

    # Ẏ is defined in the run function

    'Ż': Overload(1, {
        Number: num_length_range_0,
        Iterable: length_range_0
    }, 0, ('zero_length_range',)),

    ',': Overload(2, {
        (Any[0], Any[0]): pair
    }, 0, ('pair',)),

    '⁺': Overload(1, {
        Number: increment,
        str: str_increment
    }, 1, ('increment', 'incr', 'plus_one')),

    '⁻': Overload(1, {
        Number: decrement,
        str: str_decrement
    }, 1, ('decrement', 'decr', 'minus_one')),

    '⁼': Overload(2, {
        (Any[0], Any[0]): equals
    }, 0, ('exactly_equal', 'exactly_equals')),

    '+': Overload(2, {
        (Number[0], Number[0]): add,
        (str, Number[0]): string_add,
        (Number[0], str): string_add,
        (str, str): string_add
    }, 2, ('add', 'plus')),

    '-': Overload(2, {
        (Number[0], Number[0]): subtract,
        (str, Number[0]): slice_start,
        (Number[0], str): slice_end,
        (str, str): strip
    }, 2, ('subtract', 'minus', 'strip')),

    '×': Overload(2, {
        (Number[0], Number[0]): multiply,
        (str, Number[0]): multiply,
        (Number[0], str): multiply,
        (str, str): string_cartesian_product
    }, 2, ('multiply', 'times')),

    '/': Overload(2, {
        (Number[0], Number[0]): divide,
        (str, Number[0]): split1,
        (Number[0], str): split2,
        (str, str): split3
    }, 2, ('divide', 'split')),

    '*': Overload(2, {
        (Number[0], Number[0]): exponentiate,
        (str, Number[0]): append_first1,
        (Number[0], str): append_first2,
        (str, str): regex_findall
    }, 2, ('exponentiate', 'power', 'pow', 'regex', 'findall')),

    '%': Overload(2, {
        (Number[0], Number[0]): modulo,
        (str, Number[0]): swapped_format,
        (Number[0], str): str_format,
        (str, str): str_format
    }, 2, ('modulo', 'mod', 'format')),

    '÷': Overload(2, {
        (Number[0], Number[0]): integer_divide,
        (str, Number[0]): first_split1,
        (Number[0], str): first_split2,
        (str, str): first_split3
    }, 2, ('floor_divide', 'integer_divide')),

    '_': Overload(2, {
        (Number[0], Number[0]): swapped_subtract,
        (str, Number[0]): slice_end2,
        (Number[0], str): slice_start2,
        (str, str): strip2
    }, 2, ('swapped_subtract', 'swapped_minus', 'swapped_strip')),

    '\\': Overload(2, {
        (Number[0], Number[0]): swapped_divide,
        (str, Number[0]): split1,
        (Number[0], str): split2,
        (str, str): split4
    }, 2, ('swapped_divide', 'swapped_split')),

    '@': Overload(2, {
        (Number[0], Number[0]): swapped_exponentiate,
        (str, Number[0]): prepend_last1,
        (Number[0], str): prepend_last2,
        (str, str): swapped_regex_findall
    }, 2, ('swapped_exponentiate', 'swapped_pow', 'swapped_findall', 'swapped_regex')),

    'Œ': Overload(2, {
        (Number[0], Number[0]): swapped_modulo,
        (str, Number[0]): swapped_format,
        (Number[0], str): str_format,
        (str, str): str_format2
    }, 2, ('swapped_modulo', 'swapped_mod', 'swapped_format')),

    '¦': Overload(2, {
        (Number[0], Number[0]): swapped_integer_divide,
        (str, Number[0]): last_split1,
        (Number[0], str): last_split2,
        (str, str): last_split3
    }, 2, ('swapped_floor_divide', 'swapped_integer_divide')),

    '=': Overload(2, {
        (Any[0], Any[0]): equals
    }, 2, ('equal', 'equals')),

    '<': Overload(2, {
        (Number[0], Number[0]): less_than,
        (str, Number[0]): ord_less_than2,
        (Number[0], str): ord_less_than1,
        (str, str): less_than
    }, 2, ('less_than',)),

    '>': Overload(2, {
        (Number[0], Number[0]): greater_than,
        (str, Number[0]): ord_greater_than2,
        (Number[0], str): ord_greater_than1,
        (str, str): greater_than
    }, 2, ('greater_than',)),

    '©': Overload(2, {
        (Number[0], Number[0]): less_than_or_equal_to,
        (str, Number[0]): ord_less_than_or_equal_to2,
        (Number[0], str): ord_less_than_or_equal_to1,
        (str, str): less_than_or_equal_to
    }, 2, ('less_than_or_equal_to',)),

    '®': Overload(2, {
        (Number[0], Number[0]): greater_than_or_equal_to,
        (str, Number[0]): ord_greater_than_or_equal_to2,
        (Number[0], str): ord_greater_than_or_equal_to1,
        (str, str): greater_than_or_equal_to
    }, 2, ('greater_than_or_equal_to',)),

    '&': Overload(2, {
        (Any[0], Any[0]): logical_and
    }, 2, ('and', 'logical_and')),

    '|': Overload(2, {
        (Any[0], Any[0]): logical_or
    }, 2, ('or', 'logical_or')),

    '^': Overload(2, {
        (Any[0], Any[0]): logical_xor
    }, 2, ('xor', 'logical_xor')),

    '~': Overload(1, {
        Any: logical_not
    }, 1, ('not', 'logical_not')),

    'ð': Overload(0, {
        Any: (lambda: ' ')
    }, 0, ('space',)),

    'Ð': Overload(1, {
        Any: triplicate
    }, 0, ('triplicate',)),

    'ȧ': Overload(3, {
        (Any[0], Number[0], Iterable[0]): assign,
        (Any[0], Iterable[0], Number[0]): swapped_assign,
        (Any[0], list, Iterable[0]): vectorised_assign,
        (Any[0], str, Iterable[0]): length_assign,
        (Any[0], Number[0], Number[0]): num_assign
    }, 0, ('assign',)),

    'ḃ': Overload(1, {
        Number: to_binary,
        str: ord_bin
    }, 1, ('bin', 'binary', 'to_binary')),

    'ċ': Overload(2, {
        (Number[0], Iterable[0]): combinations1,
        (Iterable[0], Number[0]): combinations2,
        (Number[0], Number[0]): combinations3,
        (Iterable[0], Iterable[0]): set_union
    }, 0, ('combinations', 'union')),

    'ḋ': Overload(2, {
        (Number[0], Number[0]): divmod2,
        (Number[0], list): from_list_of_digits,
        (list, Number[0]): from_list_of_digits_2,
        (Any[0], Any[0]): pass_
    }, 0, ('divmod', 'from_digits')),

    'ė': Overload(1, {
        Number: exclusive_one_range,
        Iterable: length_range
    }, 0, ('length_range', 'exclusive_one_range')),

    'ḟ': Overload(1, {
        Number: prime_factor_exponents,
        str: str.title
    }, 1, ('prime_factor_exponents', 'title', 'title_case')),

    'ġ': Overload(1, {
        Number: consecutive_digits,
        Iterable: group_consecutive
    }, 0, ('group_consecutive',)),

    'ḣ': Overload(1, {
        Number: num_head_remove,
        Iterable: head_remove
    }, 0, ('head_remove',)),

    'ŀ': Overload(1, {
        Number: digits_lcm,
        str: ords_lcm,
        list: lcm
    }, 0, ('lcm', 'lowest_common_multiple')),

    'ṁ': Overload(1, {
        Number: num_median,
        str: ord_median,
        list: median
    }, 0, ('median',)),

    # ṅ is defined in the run function

    'ȯ': Overload(2, {
        (Number[0], Number[0]): round_decimal_places,
        (Number[0], Iterable[0]): range_intersection1,
        (Iterable[0], Number[0]): range_intersection2,
        (Iterable[0], Iterable[0]): set_intersection
    }, 0, ('set_intersection', 'round_dps')),

    'ṗ': Overload(2, {
        (Number[0], Iterable[0]): permutations1,
        (Iterable[0], Number[0]): permutations2,
        (Number[0], Number[0]): permutations3,
        (Iterable[0], Iterable[0]): set_difference
    }, 0, ('permutations', 'set_difference')),

    'ṙ': Overload(3, {
        (Any[0], Number[0], Any[0]): rjust1,
        (Number[0], Any[0], Any[0]): rjust2,
        (Any[0], Any[0], Any[0]): rjust3
    }, 3, ('rjust', 'right_justify')),

    'ṡ': Overload(1, {
        Number: numcumsum,
        str: prefixes,
        list: cumsum
    }, 0, ('cumsum', 'cumulative_sums')),

    'ṫ': Overload(1, {
        Number: num_tail_remove,
        Iterable: tail_remove
    }, 0, ('tail_remove',)),

    'ẇ': Overload(2, {
        (Number[0], Number[0]): chunk5,
        (Number[0], Iterable[0]): chunk3,
        (Iterable[0], Number[0]): chunk4,
        (Iterable[0], Iterable[0]): chunk6
    }, 0, ('chunk_wrap', 'split_chunk')),

    'ż': Overload(1, {
        Number: num_length_range,
        Iterable: length_range_no_pop
    }, 0, ('length_range_no_pop',)),

    'Ạ': Overload(0, {
        Any: (lambda: string.ascii_lowercase)
    }, 0, ('alphabet', 'lowercase_alphabet')),

    'Ḅ': Overload(2, {
        (int, int): ntbs,
        (str, int): character_multiply1,
        (int, str): character_multiply2,
        (str, str): regex_split,
        (float, Any): pass_
    }, 2, ('to_base_string', 'character_multiply', 'regex_split')),

    'Ḍ': Overload(1, {
        Any: double
    }, 1, ('double',)),

    'Ẹ': Overload(2, {
        (Any[0], Any[0]): exactly_not_equal
    }, 0, ('exactly_not_equal', 'exactly_inequal')),

    'Ḥ': Overload(1, {
        Number: to_hex,
        str: ord_hex
    }, 1, ('hex', 'hexadecimal', 'to_hexadecimal')),

    'Ị': Overload(1, {
        Number: reciprocal,
        str: remove_non_alphabets
    }, 1, ('reciprocal', 'inverse', 'filter_isalpha', 'only_alphabetic')),

    'Ḳ': Overload(1, {
        Number: num_bifurcate,
        Iterable: bifurcate
    }, 0, ('bifurcate',)),

    'Ḷ': Overload(2, {
        (Number[0], Number[0]): dyadic_lcm,
        (str, Number[0]): ordinal_lcm1,
        (Number[0], str): ordinal_lcm2,
        (str, str): ordinal_lcm3
    }, 2, ('dyadic_lcm',)),

    'Ṃ': Overload(1, {
        Number: num_mirror,
        Iterable: mirror
    }, 0, ('mirror',)),

    'Ṇ': Overload(3, {
        (list, list, (int, float, str)): str_transliterate,
        (list, list, list): list_transliterate,
        (list, (int, float, str), (int, float, str)): str_transliterate_overload_1,
        (list, (int, float, str), list): list_transliterate_overload_1,
        ((int, float, str), list, (int, float, str)): str_transliterate_overload_2,
        ((int, float, str), list, list): list_transliterate_overload_2,
        ((int, float, str), (int, float, str), (int, float, str)): str_transliterate_overload_3,
        ((int, float, str), (int, float, str), list): list_transliterate_overload_3
    }, 0, ('transliterate',)),

    'Ọ': Overload(2, {
        (Any[0], Any[0]): logical_or
    }, 0, ('non_vectorised_or', 'non_vectorised_logical_or')),

    'Ṛ': Overload(2, {
        (Number[0], Iterable[0]): combinations_with_replacement1,
        (Iterable[0], Number[0]): combinations_with_replacement2,
        (Number[0], Number[0]): combinations_with_replacement3,
        (Iterable[0], Iterable[0]): combinations_with_replacement4
    }, 0, ('combinations_with_replacement',)),

    'Ṣ': Overload(1, {
        Number: digit_deltas,
        str: ord_deltas,
        list: deltas
    }, 0, ('deltas', 'forward_differences')),

    'Ṭ': Overload(1, {
        Number: digits_wrap,
        str: list_wrap,
        list: transpose
    }, 0, ('transpose',)),

    'Ụ': Overload(1, {
        Number: num_rotate_left_once,
        Iterable: rotate_left_once
    }, 0, ('rotate_left_once',)),

    'Ṿ': Overload(1, {
        Number: num_rotate_right_once,
        Iterable: rotate_right_once
    }, 0, ('rotate_right_once',)),

    'Ẉ': Overload(2, {
        (Number[0], Number[0]): num_uninterleave2,
        (Number[0], Iterable[0]): uninterleave2,
        (Iterable[0], Number[0]): swapped_uninterleave,
        (Iterable[0], Iterable[0]): len_uninterleave
    }, 0, ('uninterleave_into_pieces',)),

    'Ỵ': Overload(2, {
        (Number[0], Number[0]): left_bit_shift,
        (Number[0], Iterable[0]): rotate_left,
        (Iterable[0], Number[0]): swapped_rotate_left,
        (Iterable[0], Iterable[0]): len_rotate_left
    }, 0, ('rotate_left', 'left_bit_shift')),

    'Ẓ': Overload(2, {
        (Number[0], Number[0]): right_bit_shift,
        (Number[0], Iterable[0]): rotate_right,
        (Iterable[0], Number[0]): swapped_rotate_right,
        (Iterable[0], Iterable[0]): len_rotate_right
    }, 0, ('rotate_right', 'right_bit_shift')),

    'Ä': Overload(1, {
        Number: num_to_alphabet,
        str: alphabet_to_num
    }, 1, ('num_to_alphabet', 'alphabet_to_num')),

    '¶': Overload(0, {
        Any: (lambda: '\n')
    }, 0, ('newline',)),

    '¬': Overload(1, {
        Any: logical_not
    }, 0, ('non_vectorised_not', 'non_vectorised_logical_not')),

    '§': Overload(2, {
        (Any[0], Any[0]): logical_and
    }, 0, ('non_vectorised_and', 'non_vectorised_logical_and')),

    '½': Overload(1, {
        Number: halve,
        str: chunk_halve
    }, 1, ('halve',)),

    'Ƈ': Overload(2, {
        (Number[0], Number[0]): nPr,
        (Any, str): string_contains,
        (Any, list): list_contains
    }, 0, ('npr', 'contains')),

    'Ɗ': Overload(2, {
        (Number[0], (int, float, str)): str_remove_at_index,
        (Number[0], list): list_remove_at_index,
        (str, Number[0]): swapped_str_remove_at_index,
        (list, Number[0]): swapped_list_remove_at_index,
        (Any[0], list): list_rmv,
        (list, Any[0]): swapped_list_rmv,
        (Any[0], str): str_rmv
    }, 0, ('drop_at_index', 'remove_at_index')),

    'Ƒ': Overload(1, {
        Number: unique_prime_factors,
        Iterable: substrings
    }, 0, ('unique_prime_factors', 'sublists')),

    'Ɠ': Overload(0, {
        Any: (lambda: 16)
    }, 0, ('sixteen',)),

    'Ɱ': Overload(1, {
        Number: num_palindromise,
        Iterable: palindromise
    }, 0, ('palindromise',)),

    'Ɲ': Overload(1, {
        Number: integer_partitions,
        Iterable: newline_join
    }, 0, ('integer_partitions', 'newline_join')),

    'Ƥ': Overload(2, {
        (Any[0], list): prepend,
        (list, Any[0]): swapped_prepend,
        (Any[0], Any[0]): prepend2
    }, 0, ('prepend',)),

    'Ƭ': Overload(2, {
        (Any[0], list): transpose_with_filler,
        (list, Any[0]): swapped_transpose_with_filler,
        (Number[0], Any[0]): digits_wrap_2,
        (str, Any[0]): list_wrap_2
    }, 0, ('transpose_with_filler',)),

    'ɓ': Overload(1, {
        Any: boolify
    }, 1, ('boolify',)),

    'ƈ': Overload(1, {
        Number: num_counts,
        Iterable: counts
    }, 0, ('counts',)),

    'ɗ': Overload(1, {
        Number: parity,
        str: second_half
    }, 1, ('parity', 'second_half')),

    'ƒ': Overload(1, {
        Number: num_prefixes,
        Iterable: prefixes
    }, 0, ('prefixes',)),

    'ɠ': Overload(0, {
        Any: (lambda: 256)
    }, 0, ('two_fifty_six',)),

    'ɦ': Overload(0, {
        Any: (lambda: 100)
    }, 0, ('hundred', 'one_hundred')),

    'ƙ': Overload(1, {
        Number: increment_twice,
        Iterable: grade_up
    }, 0, ('grade_up', 'increment_twice')),

    'ɱ': Overload(2, {
        (Number[0], Number[0]): num_head_slice,
        (Number[0], Iterable[0]): head_slice,
        (Iterable[0], Number[0]): swapped_head_slice,
        (Iterable[0], Iterable[0]): len_head_slice
    }, 0, ('head_slice',)),

    'ɲ': Overload(2, {
        (Number[0], Number[0]): num_tail_slice,
        (Number[0], Iterable[0]): tail_slice,
        (Iterable[0], Number[0]): swapped_tail_slice,
        (Iterable[0], Iterable[0]): len_tail_slice
    }, 0, ('tail_slice',)),

    'ƥ': Overload(1, {
        Any: pop_
    }, 0, ('pop',)),

    'ʠ': Overload(1, {
        Number: num_powerset,
        Iterable: powerset
    }, 0, ('powerset',)),

    'ɼ': Overload(1, {
        Number: randint,
        Iterable: randchoice
    }, 0, ('random', 'random_choice', 'randint')),

    'ʂ': Overload(1, {
        Number: sign,
        str: sentence_case,
        list: sum_each
    }, 0, ('sign', 'sentence_case', 'sum_each')),

    'ƭ': Overload(1, {
        Number: square_root,
        str: every_second_item
    }, 1, ('square_root', 'every_second_item')),

    '²': Overload(1, {
        Number: square,
        str: chunk_wrap_2
    }, 1, ('square',)),

    '³': Overload(1, {
        Number: cube,
        str: chunk_wrap_3
    }, 1, ('cube',)),

    '⁴': Overload(1, {
        Number: fourth,
        str: chunk_wrap_4
    }, 1, ('fourth',)),

    '⁵': Overload(1, {
        Number: fifth,
        str: chunk_wrap_5
    }, 1, ('fifth',)),

    'ạ': Overload(1, {
        Number: num_all_equal,
        Iterable: all_equal
    }, 0, ('all_equal',)),

    'ḅ': Overload(1, {
        Any: equals_one
    }, 1, ('equals_one',)),

    'ḍ': Overload(2, {
        (Number[0], Number[0]): range_ssd1,
        (Number[0], Iterable[0]): range_ssd2,
        (Iterable[0], Number[0]): range_ssd3,
        (Iterable[0], Iterable[0]): symmetric_set_difference
    }, 0, ('symmetric_set_difference',)),

    'ẹ': Overload(1, {
        Number: num_enclose,
        Iterable: enclose
    }, 0, ('enclose',)),

    'ḥ': Overload(2, {
        (list, list): add,
        (list, Any[0]): add_digits1,
        (Any[0], list): add_digits2,
        (Any[0], Any[0]): add_digits3
    }, 0, ('concatenate', 'merge')),

    'ị': Overload(2, {
        (Number[0], Number[0]): corresponding_digit_filter_1,
        (Number[0], Iterable[0]): corresponding_digit_filter_2,
        (Iterable[0], Number[0]): corresponding_digit_filter_3,
        (Iterable[0], Iterable[0]): corresponding_filter
    }, 0, ('corresponding_filter',)),

    'ḳ': Overload(1, {
        Number: num_sort_uniquify,
        str: str_sort_uniquify,
        list: list_sort_uniquify
    }, 0, ('sorted_uniquify',)),

    'ḷ': Overload(1, {
        Number: complement,
        str: islower,
        list: length_of_each
    }, 0, ('length_of_each', 'complement', 'is_lower')),

    'ṃ': Overload(1, {
        Number: math.ceil,
        str: is_vowel,
        list: reverse_each
    }, 0, ('ceil', 'is_vowel', 'reverse_each')),

    'ọ': Overload(2, {
        (Number[0], list): multiply,
        (list, Number[0]): multiply,
        (Number[0], Any[0]): digits_repeat1,
        (Any[0], Number[0]): digits_repeat2,
        (list, (int, float, str)): length_repeat1,
        (Any[0], Any[0]): length_repeat2
    }, 0, ('repeat',)),

    'ṛ': Overload(2, {
        (Number[0], Number[0]): cmp,
        (Number[0], str): zfill1,
        (str, Number[0]): zfill2,
        (str, str): surround
    }, 2, ('compare', 'cmp', 'zfill', 'surround')),

    'ṣ': Overload(1, {
        Number: spaces,
        Iterable: suffixes
    }, 0, ('suffixes', 'spaces')),

    'ṭ': Overload(3, {
        (Any[0], Any[0], Any[0]): triple_swap
    }, 0, ('triple_swap',)),

    'ẉ': Overload(2, {
        (Number[0], Number[0]): range_cartesian_power,
        (Number[0], Iterable[0]): cartesian_power,
        (Iterable[0], Number[0]): swapped_cartesian_power,
        (Iterable[0], Iterable[0]): cartesian_product
    }, 0, ('cartesian_power',)),

    'ỵ': Overload(2, {
        (Number[0], Number[0]): num_head_remove_2,
        (Number[0], Iterable[0]): head_remove_2,
        (Iterable[0], Number[0]): swapped_head_remove_2,
        (Iterable[0], Iterable[0]): len_head_remove_2
    }, 0, ('head_remove_many',)),

    'ẓ': Overload(2, {
        (Number[0], Number[0]): num_tail_remove_2,
        (Number[0], Iterable[0]): tail_remove_2,
        (Iterable[0], Number[0]): swapped_tail_remove_2,
        (Iterable[0], Iterable[0]): len_tail_remove_2
    }, 0, ('tail_remove_many',)),

}


def get_a_function(command):
    if len(command) == 1:
        return commands.get(command, Void)
    # Add digraph support here
    return Void



from thunno2 import dictionary
from thunno2.commands import *
from thunno2.lexer import tokenise
import string
import copy
import sys

"""The main interpreter. To run a Thunno 2 program, use the run function."""

vars_dict = {
    'x': 1,   # x defaults to 1
    'y': 2,   # y defaults to 2
    'ga': []  # global array defaults to []
}

'''RUN FUNCTION'''


def run(code, *, n, iteration_index):
    # while ctx.index < len(code):
    for chars, desc, info in code:
        if desc == 'command' or desc == 'digraph':
            if info != Void:
                values = info()
                for value in values:
                    ctx.stack.push(value)
        elif desc in (
            'number', 'string', 'one character', 'two characters', 'three characters', 'list'
        ):
            ctx.stack.push(info)
        elif desc == 'lowercase alphabetic compression':
            base255_number = decompress(info, '“')
            decompressed_string = to_custom_base_string(' ' + string.ascii_lowercase, base255_number)
            ctx.stack.push(decompressed_string)
        elif desc == 'title case alphabetic compression':
            base255_number = decompress(info, '”')
            decompressed_string = to_custom_base_string(' ' + string.ascii_lowercase, base255_number)
            ctx.stack.push(decompressed_string.title())
        elif desc == 'lowercase dictionary compression':
            lst = []
            i = 0
            while i < len(info):
                c = info[i]
                if c in dictionary.dictionary_codepage:
                    try:
                        i += 1
                        lst.append(dictionary.dictionary_decompress_string(c + info[i]))
                    except:
                        pass
                elif c == '\\':
                    try:
                        i += 1
                        lst.append(info[i])
                    except:
                        lst.append('\\')
                else:
                    lst.append(c)
                i += 1
            ctx.stack.push(''.join(lst))
        elif desc == 'title case dictionary compression':
            lst = []
            i = 0
            while i < len(info):
                c = info[i]
                if c in dictionary.dictionary_codepage:
                    try:
                        i += 1
                        lst.append(dictionary.dictionary_decompress_string(c + info[i]).title())
                    except:
                        pass
                elif c == '\\':
                    try:
                        i += 1
                        lst.append(info[i])
                    except:
                        lst.append('\\')
                else:
                    lst.append(c)
                i += 1
            ctx.stack.push(''.join(lst))
        elif desc == 'compressed number' or desc == 'small compressed number':
            base255_number = decompress(info, '»')
            ctx.stack.push(base255_number)
        elif desc == 'compressed list':
            base255_number = decompress(info, '¿')
            decompressed_string = to_custom_base_string('][0123456789-.,', base255_number)
            try:
                e = eval(decompressed_string)
                if isinstance(e, tuple):
                    ctx.stack.push(list(e))
                else:
                    ctx.stack.push(e)
            except:
                ctx.stack.push(decompressed_string)
        elif desc == 'variable get':
            ctx.stack.push(vars_dict.get(info, 0))
        elif desc == 'variable set':
            a = next(ctx.stack.rmv(1))
            vars_dict[info] = a
        elif desc == 'single function map':
            a = next(ctx.stack.rmv(1))
            func = info
            if func != Void:
                x = listify(a)
                r = []
                old_stack = Stack(copy.deepcopy(list(ctx.stack).copy()))
                for i in x:
                    ctx.stack = Stack(copy.deepcopy(list(old_stack).copy()))
                    ctx.stack.push(i)
                    for j in func():
                        r.append(j)
                ctx.stack.push(r)
        elif desc == 'outer product':
            a, b = ctx.stack.rmv(2)
            func = info
            if func != Void:
                x = listify(a)
                r = []
                old_stack = Stack(copy.deepcopy(list(ctx.stack).copy()))
                if isinstance(a, list):
                    if isinstance(b, list):
                        for j in b:
                            r.append([])
                            for i in a:
                                ctx.stack = Stack([i, j] + copy.deepcopy(list(old_stack).copy()))
                                for k in func():
                                    r[-1].append(k)
                    else:
                        for i in a:
                            ctx.stack = Stack([i, b] + copy.deepcopy(list(old_stack).copy()))
                            for k in func():
                                r.append(k)
                else:
                    if isinstance(b, list):
                        for i in b:
                            ctx.stack = Stack([a, i] + copy.deepcopy(list(old_stack).copy()))
                            for k in func():
                                r.append(k)
                    else:
                        ctx.stack.push(b)
                        ctx.stack.push(a)
                        k = func()
                        if k:
                            r = k[-1]
                ctx.stack.push(r)
        elif desc == 'single function filter':
            a = next(ctx.stack.rmv(1))
            func = info
            if func != Void:
                x = listify(a)
                r = []
                old_stack = Stack(copy.deepcopy(list(ctx.stack).copy()))
                for i in x:
                    ctx.stack = Stack(copy.deepcopy(list(old_stack).copy()))
                    ctx.stack.push(i)
                    f = func()
                    if not f:
                        continue
                    if f[-1]:
                        r.append(i)
                ctx.stack.push(r)
        elif desc == 'single function sort by':
            a = next(ctx.stack.rmv(1))
            func = info
            if func != Void:
                x = listify(a)
                sort_by = []
                old_stack = Stack(copy.deepcopy(list(ctx.stack).copy()))
                for i in x:
                    ctx.stack = Stack(copy.deepcopy(list(old_stack).copy()))
                    ctx.stack.push(i)
                    f = func()
                    if not f:
                        sort_by.append((i, i))
                    else:
                        sort_by.append((i, f[-1]))
                try:
                    sorted_list = sorted(sort_by, key=lambda t: t[-1])
                    ctx.stack.push([p for p, q in sorted_list])
                except:
                    ctx.stack.push(x)
        elif desc == 'single function group by':
            a = next(ctx.stack.rmv(1))
            func = info
            if func != Void:
                x = listify(a)
                group_by = []
                old_stack = Stack(copy.deepcopy(list(ctx.stack).copy()))
                for i in x:
                    ctx.stack = Stack(copy.deepcopy(list(old_stack).copy()))
                    ctx.stack.push(i)
                    f = func()
                    if not f:
                        group_by.append((i, i))
                    else:
                        group_by.append((i, f[-1]))
                try:
                    d = []
                    for val, key in group_by:
                        for k, (i, j) in enumerate(d):
                            if key == i:
                                d[k][1].append(val)
                                break
                        else:
                            d.append((key, [val]))
                    ctx.stack.push([q for p, q in d])
                except:
                    ctx.stack.push(x)
        elif desc == 'context variable':
            ctx.stack.push(n)
        elif desc == 'iteration index':
            ctx.stack.push(iteration_index)
        elif desc == 'get x':
            ctx.stack.push(vars_dict.get('x', 1))
        elif desc == 'get y':
            ctx.stack.push(vars_dict.get('y', 2))
        elif desc == 'set x':
            a = next(ctx.stack.rmv(1))
            vars_dict['x'] = a
        elif desc == 'set y':
            a = next(ctx.stack.rmv(1))
            vars_dict['y'] = a
        elif desc == 'set x without popping':
            a = (ctx.stack.copy() + ctx.other_il + [0])[0]
            vars_dict['x'] = a
        elif desc == 'set y without popping':
            a = (ctx.stack.copy() + ctx.other_il + [0])[0]
            vars_dict['y'] = a
        elif desc == 'increment x':
            try:
                vars_dict['x'] = vars_dict.get('x', 1) + 1
            except:
                pass
        elif desc == 'increment y':
            try:
                vars_dict['y'] = vars_dict.get('y', 2) + 1
            except:
                pass
        elif desc == 'get global array':
            ctx.stack.push(vars_dict.get('ga', []))
        elif desc == 'add to global array':
            a = next(ctx.stack.rmv(1))
            ga = vars_dict.get('ga', [])
            if not isinstance(ga, list):
                vars_dict['ga'] = [ga, a]
            vars_dict['ga'] = ga + [a]
        elif desc == 'stack':
            ctx.stack.push(list(ctx.stack).copy())
        elif desc == 'constant':
            ctx.stack.push(info)
        elif desc == 'codepage compression':
            ctx.stack.push(info)
        elif desc == 'quit':
            raise TerminateProgramException()  # This will hopefully get caught and ignored
        elif desc == 'next input':
            if ctx.other_il:
                ctx.stack.push(ctx.other_il[0])
                ctx.other_il = ctx.other_il[1:] + [ctx.other_il[0]]
        elif desc == 'input list':
            ctx.stack.push(ctx.og_input_list)
        elif desc == 'first input':
            try:
                ctx.stack.push(ctx.og_input_list[0])
            except:
                ctx.stack.push(ctx.og_input_list)
        elif desc == 'second input':
            try:
                ctx.stack.push(ctx.og_input_list[1])
            except:
                ctx.stack.push(ctx.og_input_list)
        elif desc == 'third input':
            try:
                ctx.stack.push(ctx.og_input_list[2])
            except:
                ctx.stack.push(ctx.og_input_list)
        elif desc == 'third last input':
            try:
                ctx.stack.push(ctx.og_input_list[-3])
            except:
                ctx.stack.push(ctx.og_input_list)
        elif desc == 'second last input':
            try:
                ctx.stack.push(ctx.og_input_list[-2])
            except:
                ctx.stack.push(ctx.og_input_list)
        elif desc == 'last input':
            try:
                ctx.stack.push(ctx.og_input_list[-1])
            except:
                ctx.stack.push(ctx.og_input_list)
        elif desc == 'print':
            print(next(ctx.stack.rmv(1)))
            ctx.implicit_print = False
        elif desc == 'print without newline':
            print(next(ctx.stack.rmv(1)), end='')
            ctx.implicit_print = False
        elif desc == 'print without popping':
            print((ctx.stack.copy() + ctx.other_il + [0])[0])
        elif desc == 'map':
            a = next(ctx.stack.rmv(1))
            x = listify(a)
            r = []
            old_stack = Stack(copy.deepcopy(list(ctx.stack).copy()))
            for i, j in enumerate(x):
                ctx.stack = Stack([j] + copy.deepcopy(old_stack))
                run(info, n=j, iteration_index=i)
                r.append(next(ctx.stack.rmv(1)))
            ctx.stack.push(r)
        elif desc == 'filter':
            a = next(ctx.stack.rmv(1))
            x = listify(a)
            r = []
            old_stack = Stack(copy.deepcopy(list(ctx.stack).copy()))
            for i, j in enumerate(x):
                ctx.stack = Stack([j] + copy.deepcopy(old_stack))
                run(info, n=j, iteration_index=i)
                z = next(ctx.stack.rmv(1))
                if z:
                    r.append(j)
            ctx.stack.push(r)
        elif desc == 'sort by':
            a = next(ctx.stack.rmv(1))
            x = listify(a)
            r = []
            old_stack = Stack(copy.deepcopy(list(ctx.stack).copy()))
            for i, j in enumerate(x):
                ctx.stack = Stack([j] + copy.deepcopy(old_stack))
                run(info, n=j, iteration_index=i)
                z = next(ctx.stack.rmv(1))
                r.append((j, z))
            try:
                sorted_list = sorted(r, key=lambda t: t[-1])
                ctx.stack.push([p for p, q in sorted_list])
            except:
                ctx.stack.push(x)
        elif desc == 'group by':
            a = next(ctx.stack.rmv(1))
            x = listify(a)
            r = []
            old_stack = Stack(copy.deepcopy(list(ctx.stack).copy()))
            for i, j in enumerate(x):
                ctx.stack = Stack([j] + copy.deepcopy(old_stack))
                run(info, n=j, iteration_index=i)
                z = next(ctx.stack.rmv(1))
                r.append((j, z))
            try:
                d = []
                for val, key in r:
                    for k, (i, j) in enumerate(d):
                        if key == i:
                            d[k][1].append(val)
                            break
                    else:
                        d.append((key, [val]))
                ctx.stack.push([q for p, q in d])
            except:
                ctx.stack.push(x)
        elif desc == 'fixed point':
            r = [Void]
            old_stack = Stack(copy.deepcopy(list(ctx.stack).copy()))
            i = 0
            while True:
                ctx.stack = Stack(copy.deepcopy(old_stack))
                run(info, n=r[-1], iteration_index=i)
                k = (ctx.stack + ctx.other_il + [0])[0]
                print(k)
                r.append(k)
                if r[-1] == r[-2]:
                    break
                i += 1
            ctx.stack.push(r[1:])
        elif desc == 'first n integers':
            a = next(ctx.stack.rmv(1))
            try:
                x = int(a)
            except:
                x = 1
            old_stack = Stack(copy.deepcopy(list(ctx.stack).copy()))
            i = 1
            r = []
            while len(r) < x:
                ctx.stack = Stack(copy.deepcopy(old_stack))
                ctx.stack.push(i)
                run(info, n=i, iteration_index=i-1)
                k = next(ctx.stack.rmv(1))
                if k:
                    r.append(i)
                i += 1
            ctx.stack.push(r)
        elif desc == 'cumulative reduce by':
            a = next(ctx.stack.rmv(1))
            x = listify(a)
            old_stack = Stack(copy.deepcopy(list(ctx.stack).copy()))
            if x:
                r = [x.pop(0)]
                for i, j in enumerate(x):
                    ctx.stack = Stack(copy.deepcopy(old_stack))
                    ctx.stack.push(r[-1])
                    ctx.stack.push(j)
                    run(info, n=j, iteration_index=i)
                    r.append(next(ctx.stack.rmv(1)))
                ctx.stack.push(r)
            else:
                ctx.stack.push([])
        elif desc == 'for loop':
            a = next(ctx.stack.rmv(1))
            x = listify(a)
            for i, j in enumerate(x):
                if not isinstance(a, (int, float)):
                    ctx.stack.push(j)
                run(info, n=j, iteration_index=i)
        elif desc == 'while loop':
            cond, body = info
            i = 0
            while True:
                old_stack = Stack(copy.deepcopy(list(ctx.stack).copy()))
                run(cond, n=0, iteration_index=i)
                z = next(ctx.stack.rmv(1))
                ctx.stack = Stack(copy.deepcopy(old_stack))
                if not z:
                    break
                run(body, n=0, iteration_index=i)
                i += 1
        elif desc == 'if statement':
            if_true, if_false = info
            a = next(ctx.stack.rmv(1))
            if a:
                run(if_true, n=0, iteration_index=1)
            else:
                run(if_false, n=0, iteration_index=0)
        elif desc == 'execute without popping':
            if info != Void:
                values = info(pop=False)
                for value in values:
                    ctx.stack.push(value)
        elif desc == 'pair apply':
            a = next(ctx.stack.rmv(1))
            old_stack = Stack(copy.deepcopy(list(ctx.stack).copy()))
            r = []
            f1, f2 = info
            ctx.stack.push(a)
            k = f1()
            if k:
                r.append(k[-1])
            ctx.stack = Stack(copy.deepcopy(old_stack))
            ctx.stack.push(a)
            k = f2()
            if k:
                r.append(k[-1])
            ctx.stack.push(r)
        elif desc == 'two function map':
            a = next(ctx.stack.rmv(1))
            old_stack = Stack(copy.deepcopy(list(ctx.stack).copy()))
            r = []
            f1, f2 = info
            x = listify(a)
            for i in x:
                ctx.stack = Stack(copy.deepcopy(list(old_stack).copy()))
                ctx.stack.push(i)
                for j in f1():
                    ctx.stack.push(j)
                for k in f2():
                    ctx.stack.push(k)
                r.append(next(ctx.stack.rmv(1)))
            ctx.stack.push(r)
        else:
            if ctx.warnings:
                print('TRACEBACK: [UNRECOGNISED TOKEN]', file=sys.stderr)
                print(f'Got {chars!r} (tokenised to {desc!r})')
    return 0


def test(cod, inp=(), stk=(), warn=True):
    ctx.stack = Stack(stk)
    ctx.og_input_list = list(inp)
    ctx.other_il = list(inp)
    ctx.warnings = warn
    tokenised = tokenise(cod)[1]
    run(tokenised, n=0, iteration_index=0)
    print(ctx.stack)
    if ctx.implicit_print:
        print(ctx.stack[0])



"""An unordered, chaotic mess of helper functions"""
import copy
import itertools
import math
import random
import re
import string

from thunno2 import codepage


def safe_max_len(*lsts):
    m = 0
    for lst in lsts:
        if isinstance(lst, list):
            if len(lst) > m:
                m = len(lst)
    return m


def first_non_none(l):
    for i in l:
        if i is not None:
            return i


def isalpha(s):
    return int(s.isalpha())


def number_to_base_digits(n, b):
    if n == 0:
        return [0]
    digits = []
    while n:
        digits.append(int(n % b))
        n //= b
    return digits[::-1]


def ntb(b, n):
    if b == 0:
        return [n]
    if b == 1:
        return n * [0]
    return number_to_base_digits(n, b)


def number_to_base(base, num):
    num, base = int(num), int(base)
    if base <= 0:
        return str(num)
    if base == 1:
        return num * '0'
    return number_to_base_digits(num, base)


def ntbs(b, n):
    r = number_to_base_digits(n, b)
    if b <= 64:
        digits = string.digits + string.ascii_uppercase + string.ascii_lowercase + '+/'
    else:
        digits = codepage.CODEPAGE * math.ceil(b / 256)
    return ''.join(map(digits.__getitem__, r))


def to_custom_base_string(base, num):
    num = int(num)
    if len(base) == 0:
        return str(num)
    if len(base) == 1:
        return num * base
    return ''.join(map(base.__getitem__, number_to_base_digits(num, len(base))))


def length_to_base(base, s):
    base = int(base)
    if base <= 0:
        return str(len(s))
    if base == 1:
        return len(s) * '0'
    return number_to_base_digits(len(s), base)


def length_custom_base_string(str2, str1):
    num = len(str1)
    if len(str2) == 0:
        return str(num)
    if len(str2) == 1:
        return num * '0'
    return ''.join(map(str2.__getitem__, number_to_base_digits(len(str1), len(str2))))


def pass_(*args):
    return args[::-1]


def from_list_of_digits_2(num, base):
    r = 0
    for x, y in enumerate(reversed(num)):
        r += int(base) ** x * int(y)
    return r


def decompress(s, char):
    ci = codepage.CODEPAGE.index(char)
    indices = [i - (i > ci) for i in codepage.codepage_index(*s)]
    return from_list_of_digits_2(indices, 255)


def chr2(num):
    return chr(int(num))


def ord2(s):
    return [ord(c) for c in s]


def digits(num):
    return [int(d) for d in str(num) if d in '0123456789']


def is_even(num):
    return int(num % 2 == 0)


def eval2(s):
    try:
        return eval(s)
    except:
        return 0


def factors(num):
    num = int(num)
    return [i for i in range(1, num + 1) if num % i == 0]


def substrings(s):
    return [s[i:j] for i in range(len(s)) for j in range(i + 1, len(s) + 1)]


def max2(num):
    return max(int(d) for d in str(num) if d in '0123456789')


def max3(x):
    try:
        return max(x)
    except:
        return x


def from_hex(x):
    try:
        return int(str(x), 16)
    except:
        return x


def codepage_chr(num):
    return codepage.CODEPAGE[int(num) % 256]


def codepage_ord(s):
    return [*codepage.codepage_index(*s)]


def indexing_0(num, lst):
    if not lst:
        return lst
    return lst[int(num) % len(lst)]


def swapped_ind0(lst, num):
    if not lst:
        return lst
    return lst[int(num) % len(lst)]


def interleave_lst(lst1, lst2):
    if len(lst1) == len(lst2):
        r = []
        for i, j in zip(lst2, lst1):
            r.append(i)
            r.append(j)
        return r
    elif len(lst1) > len(lst2):
        r = []
        for i, j in zip(lst2, lst1):
            r.append(i)
            r.append(j)
        r.extend(lst1[len(lst2):])
        return r
    else:
        r = []
        for i, j in zip(lst2, lst1):
            r.append(i)
            r.append(j)
        r.extend(lst2[len(lst1):])
        return r


def interleave_str(s1, s2):
    if len(s1) == len(s2):
        r = ''
        for i, j in zip(s2, s1):
            r += i + j
        return r
    elif len(s1) > len(s2):
        r = ''
        for i, j in zip(s2, s1):
            r += i + j
        r += s1[len(s2):]
        return r
    else:
        r = ''
        for i, j in zip(s2, s1):
            r += i + j
        r += s2[len(s1):]
        return r


def binary_range(num1, num2):
    if num1 < num2:
        return binary_range(num2, num1)[::-1]
    return list(range(num2, num1 + 1))


def empty_join(iterable):
    if not isinstance(iterable, list):
        return str(iterable)
    return ''.join(map(empty_join, iterable))


def safe_len(x):
    try:
        return len(x)
    except:
        return 1


def lowered_range(num):
    if num < 0:
        return [-i for i in lowered_range(-num)]
    return list(range(int(num)))


def min2(num):
    return min(int(d) for d in str(num) if d in '0123456789')


def min3(x):
    try:
        return min(x)
    except:
        return x


def safe_int(x):
    try:
        return int(x)
    except:
        return str(x)


def two_power(n):
    return 2 ** n


def is_prime(n):
    i = int(n)
    return int(i >= 2 and all(i % k for k in range(2, i)))


def slice1(num, lst):
    return lst[::int(num)]


def slice2(lst, num):
    return lst[::int(num)]


def not_equal(x, y):
    return int(x != y)


def one_range(num):
    if num < 0:
        return [-i for i in one_range(-num)]
    return list(range(1, int(num) + 1))


def digit_sum(num):
    return sum(int(d) for d in str(num) if d in '0123456789')


def it_sum(lst):
    if not lst:
        return 0
    if all(isinstance(item, (int, float)) for item in lst):
        return sum(lst)
    if all(isinstance(item, list) for item in lst):
        return sum(lst[1:], lst[0])
    return ''.join(map(str, lst))


def uniquify_lst(lst):
    r = []
    for i in lst:
        if i not in r:
            r.append(i)
    return r


def uniquify_str(s):
    r = ''
    for c in s:
        if c not in r:
            r += c
    return r


def uniquify_num(num):
    return eval(uniquify_str(str(num)))


def indices_where_truthy(l):
    return [i for i, x in enumerate(l) if x]


def indices_where_truthy_num(n):
    return [i for i, x in enumerate(str(n)) if x in '123456789']


def rot_13(s):
    r = ''
    for c in s:
        if ord('A') <= ord(c.upper()) <= ord('M'):
            r += chr(ord(c) + 13)
        elif ord('N') <= ord(c.upper()) <= ord('Z'):
            r += chr(ord(c) - 13)
        else:
            r += c
    return r


def wrap(x):
    return [x]


def zip2(l1, l2):
    return list(map(list, zip(l2, l1)))


def range_zip1(n1, n2):
    n1 = int(n1)
    return zip2(one_range(n1), [n2] * abs(n1))


def range_zip2(n, l):
    return zip2(one_range(int(n)), l)


def range_zip3(l, n):
    return zip2(l, one_range(int(n)))


def append(x, l):
    return l + [x]


def safe_index(x, y, e=-1):
    if y in x:
        return x.index(y)
    return e


def convert_from_base(base, num):
    if base <= 0:
        return num
    elif base == 1:
        return len(str(num))
    if base <= 64:
        digits = string.digits + string.ascii_uppercase + string.ascii_lowercase + '+/'
    else:
        digits = codepage.CODEPAGE * math.ceil(base / 256)
    return from_list_of_digits_2(list(map((lambda c: safe_index(digits, c, 0)), str(num))), base)


def convert_from_custom_base(base, num):
    return from_list_of_digits_2(list(map((lambda c: safe_index(base, c, 0)), str(num))), len(base))


def nCr(r, n):
    n, r = int(n), int(r)
    f = math.factorial
    return f(n) // f(r) // f(n - r)


def string_count(x, y):
    if isinstance(x, list):
        return [string_count(i, y) for i in x]
    return str(y).count(str(x))


def list_count(x, y):
    if isinstance(x, list):
        return [list_count(i, y) for i in x]
    return y.count(x)


def duplicate(x):
    return x, x


def ten_power(n):
    return 10 ** n


def comma_split(s):
    return s.split(',')


def prime_factors(n):
    n = int(n)
    for i in range(2, n + 1):
        if is_prime(i) and n % i == 0:
            return [i] + prime_factors(n // i)
    return []


def case(s):
    r = []
    for c in s:
        if c.islower():
            r.append(0)
        elif c.isupper():
            r.append(1)
        else:
            r.append(-1)
    return r


def gcd(l):
    l = [int(i) for i in l if isinstance(i, (int, float)) or str(i).isdigit()]
    if not l:
        return 0
    for x in reversed(range(1, max(l))):
        if all(y % x == 0 for y in l):
            return x


def digits_gcd(n):
    return gcd([int(i) for i in str(n) if i in '0123456789'])


def ords_gcd(s):
    return gcd(list(map(ord, s)))


def head_extract(x):
    if not x:
        return x
    return x[1:], x[0]


def num_head_extract(n):
    r = ()
    for i in head_extract(str(n)):
        try:
            r += eval(i),
        except:
            r += i,
    return r


def num_ind0(x, y):
    s = str(y)[int(x) % len(str(y))]
    try:
        return int(s)
    except:
        return s


def length_ind0(x, y):
    return indexing_0(len(str(x)), y)


def vectorised_ind0(x, y):
    r = []
    for i in x:
        try:
            r.append(indexing_0(i, y))
        except:
            r.append(length_ind0(i, y))
    return r


def increment(num):
    return num + 1


def decrement(num):
    return num - 1


def str_increment(s):
    return ''.join(chr(ord(c) + 1) for c in s)


def str_decrement(s):
    return ''.join(chr(ord(c) - 1) for c in s)


def join(joiner, iterable):
    return str(joiner).join(map(str, iterable))


def num_join(j, n):
    return join(j, str(n))


def num_length(n):
    return len(str(n))


def mean(lst):
    l = []
    for i in lst:
        try:
            l.append(float(i))
        except:
            pass
    if not l:
        return 0
    return sum(l) / len(l)


def num_mean(n):
    return mean([int(i) for i in str(n) if i in '0123456789'])


def ord_mean(s):
    return mean(map(ord, s))


def str_rmv(x, y):
    return str(y).replace(str(x), '')


def num_rmv(x, y):
    try:
        return eval(str_rmv(x, y))
    except:
        return str_rmv(x, y)


def list_rmv(x, y):
    z = y.copy()
    while x in z:
        z.remove(x)
    return z


def swapped_list_rmv(x, y):
    return list_rmv(y, x)


def digit_product(num):
    return math.prod([int(i) for i in str(num) if i in '0123456789'])


def product(lst):
    l = []
    for i in lst:
        if isinstance(i, (int, float)):
            l.append(i)
        else:
            try:
                l.append(int(i))
            except:
                try:
                    l.append(float(i))
                except:
                    pass
    if not l:
        return 0
    return math.prod(l)


def ord_product(s):
    return math.prod(list(map(ord, s)))


def swapped_append(x, y):
    return append(y, x)


def _digits(x):
    if isinstance(x, (str, list)):
        return list(x)
    return [int(i) for i in str(x) if i in '0123456789']


def append2(x, y):
    return _digits(y) + [x]


def string_replace(x, y, z):
    return str(z).replace(str(y), str(x))


def list_replace(x, y, z):
    return [x if i == y else i for i in z]


def digit_reverse(num):
    r = str(num)[::-1]
    try:
        return eval(r)
    except:
        return r


def reverse(x):
    return x[::-1]


def swap(a, b):
    return a, b


def tail_extract(x):
    if not x:
        return x
    return x[:-1], x[-1]


def num_tail_extract(n):
    r = ()
    for i in tail_extract(str(n)):
        try:
            r += eval(i),
        except:
            r += i,
    return r


def remove_whitespace(s):
    return ''.join(s.split())


def uninterleave(l):
    return [l[0::2], l[1::2]]


def num_uninterleave(n):
    return uninterleave(_digits(n))


def bool2(x):
    return int(bool(x))


def isalphanum(s):
    return int(s.isalnum())


def any2(l):
    return int(any(l))


def from_binary(x):
    try:
        return int(str(x), 2)
    except:
        return x


def add(x, y):
    return y + x


def string_add(x, y):
    return str(y) + str(x)


def subtract(x, y):
    return y - x


def slice_start(s, n):
    return s[n:]


def slice_end(n, s):
    return s[:-n]


def strip(a, b):
    return b.strip(a)


def multiply(x, y):
    if isinstance(x, (str, list)):
        return x * int(y)
    elif isinstance(y, (str, list)):
        return int(x) * y
    return y * x


def string_cartesian_product(a, b):
    return [''.join(l) for l in itertools.product(b, a)]


def divide(x, y):
    return y / x


def split2(a, b):
    if a < 1:
        return b
    if not b:
        return [b] * a
    x, y = divmod(len(b), int(a))
    k = 0
    r = []
    while k < len(b):
        r.append(b[k:k + x + (y > 0)])
        k += x + (y > 0)
        if y > 0:
            y -= 1
    return r


def split1(a, b):
    return split2(b, a)


def split3(a, b):
    return b.split(a)


def exponentiate(x, y):
    return y ** x


def append_first1(a, b):
    if not a:
        return ' ' * b
    while len(a) < b:
        a += a[0]
    return a


def append_first2(a, b):
    return append_first1(b, a)


def regex_findall(a, b):
    return list(re.findall(a, b))


def modulo(x, y):
    return y % x


def str_format(a, b):
    return str(b).replace('%', str(a))


def swapped_format(a, b):
    return str_format(b, a)


def integer_divide(x, y):
    return int(y // x)


def first_split1(a, b):
    s = split1(a, b)
    if s:
        return s[0]
    return s


def first_split2(a, b):
    s = split2(a, b)
    if s:
        return s[0]
    return s


def first_split3(a, b):
    s = split3(a, b)
    if s:
        return s[0]
    return s


def slice_start2(n, s):
    return s[n:]


def slice_end2(s, n):
    return s[:-n]


def swapped_subtract(x, y):
    return x - y


def strip2(a, b):
    return a.strip(b)


def split4(a, b):
    return a.split(b)


def swapped_divide(x, y):
    return x / y


def swapped_integer_divide(x, y):
    return int(x // y)


def last_split1(a, b):
    s = split1(a, b)
    if s:
        return s[-1]
    return s


def last_split2(a, b):
    s = split2(a, b)
    if s:
        return s[-1]
    return s


def last_split3(a, b):
    s = split3(a, b)
    if s:
        return s[-1]
    return s


def str_format2(a, b):
    return str(a).replace('%', str(b))


def swapped_modulo(x, y):
    return x % y


def swapped_exponentiate(x, y):
    return x ** y


def prepend_last1(a, b):
    if not a:
        return ' ' * b
    while len(a) < b:
        a = a[-1] + a
    return a


def prepend_last2(a, b):
    return append_first1(b, a)


def swapped_regex_findall(a, b):
    return list(re.findall(b, a))


def equals(x, y):
    return int(y == x)


def less_than(x, y):
    return int(y < x)


def ord_less_than1(n, s):
    return [int(ord(x) < n) for x in s]


def ord_less_than2(s, n):
    return [int(n < ord(x)) for x in s]


def greater_than(x, y):
    return int(y > x)


def ord_greater_than1(n, s):
    return [int(ord(x) > n) for x in s]


def ord_greater_than2(s, n):
    return [int(n > ord(x)) for x in s]


def less_than_or_equal_to(x, y):
    return int(y <= x)


def ord_less_than_or_equal_to1(n, s):
    return [int(ord(x) <= n) for x in s]


def ord_less_than_or_equal_to2(s, n):
    return [int(n <= ord(x)) for x in s]


def greater_than_or_equal_to(x, y):
    return int(y >= x)


def ord_greater_than_or_equal_to1(n, s):
    return [int(ord(x) >= n) for x in s]


def ord_greater_than_or_equal_to2(s, n):
    return [int(n >= ord(x)) for x in s]


def logical_and(x, y):
    return y and x


def logical_or(x, y):
    return y or x


def logical_xor(x, y):
    return int(bool(y) ^ bool(x))


def logical_not(x):
    return int(not x)


def pair(x, y):
    return [y, x]


def is_divisible(x, y):
    try:
        return int(y % x == 0)
    except:
        return 0


def length_divisible1(s, n):
    return is_divisible(len(s), n)


def length_divisible2(n, s):
    return length_divisible1(s, n)


def length_divisible3(a, b):
    return is_divisible(len(a), len(b))


def zero_enumerate(l):
    return list(map(list, enumerate(l)))


def inclusive_zero_range(n):
    if n >= 0:
        return list(range(int(n) + 1))
    return [-i for i in range(abs(int(n)) + 1)]


def recursive_flatten(x):
    if isinstance(x, list):
        r = []
        for i in x:
            a = recursive_flatten(i)
            if isinstance(a, list):
                r += a
            else:
                r += a,
        return r
    return x


def dyadic_gcd(a, b):
    try:
        return math.gcd(abs(int(a)), abs(int(b)))
    except:
        return 1


def ordinal_gcd1(a, b):
    return [dyadic_gcd(ord(c), b) for c in a]


def ordinal_gcd2(a, b):
    return ordinal_gcd1(b, a)


def longest_common_substring(x, y):
    a, b = substrings(x), substrings(y)
    try:
        return max([i for i in a if i in b], key=len)
    except:
        return ''


def num_head(x):
    return int(str(x)[0])


def head(x):
    if not x:
        return x
    return x[0]


def num_tail(x):
    return int(str(x)[-1])


def tail(x):
    if not x:
        return x
    return x[-1]


def num_ind1(x, y):
    s = str(y)[~-int(x) % len(str(y))]
    try:
        return int(s)
    except:
        return s


def indexing_1(num, lst):
    if not lst:
        return lst
    return lst[~-int(num) % len(lst)]


def swapped_ind1(lst, num):
    if not lst:
        return lst
    return lst[~-int(num) % len(lst)]


def length_ind1(x, y):
    return indexing_1(len(str(x)), y)


def vectorised_ind1(x, y):
    r = []
    for i in x:
        try:
            r.append(indexing_1(i, y))
        except:
            r.append(length_ind1(i, y))
    return r


def ljust1(x, n, y):
    return str(y).ljust(abs(int(n)) or 1, (str(x) + ' ')[0])


def ljust2(n, x, y):
    return str(y).ljust(abs(int(n)) or 1, (str(x) + ' ')[0])


def ljust3(x, y, z):
    return str(z).ljust(len(y) or 1, (str(x) + ' ')[0])


def mode(it):
    if not it:
        return it
    return max(it, key=it.count)


def num_mode(n):
    return mode(digits(n))


def negate(x):
    return -x


def rle(s):
    l = ''
    r = []
    for c in s:
        if c == l:
            r[-1] += 1
        else:
            r.append(1)
        l = c
    return r


def index_of(x, y):
    if isinstance(y, str):
        x = str(x)
    if x in y:
        return y.index(x)
    return -1


def swapped_index_of(x, y):
    return index_of(y, x)


def num_index_of(x, y):
    return index_of(x, digits(y))


def vectorised_index_of(x, y):
    r = []
    for i in x:
        if isinstance(y, (str, list)):
            r.append(index_of(i, y))
        else:
            r.append(num_index_of(i, y))


def cartesian_product(x, y):
    return list(map(list, itertools.product(y, x)))


def range_product1(n, y):
    return cartesian_product(list(one_range(n)), y)


def range_product2(x, n):
    return range_product1(n, x)


def range_product3(x, y):
    return cartesian_product(list(one_range(x)), list(one_range(y)))


def first_char_not_present(s, avoid=()):
    i = 0
    while 1:
        if chr(i) in avoid:
            continue
        if chr(i) not in s:
            return chr(i)
        i += 1


def recursive_replace(x, y, z):
    while y in z:
        z = z.replace(y, x)
    return z


def string_repr(s):
    r = repr(s)
    if r[0] == '"':
        return r
    f = first_char_not_present(r, '\\\'"')
    return '"' + recursive_replace(f, '\\\\', r[1:-1]).replace("\\'", "'").replace('"', '\\"').replace(f, '\\\\') + '"'


def list_repr(l):
    x = []
    for i in l:
        if isinstance(i, list):
            x.append(list_repr(i))
        elif isinstance(i, str):
            x.append(string_repr(i))
        else:
            x.append(str(i))
    return '[' + ', '.join(x) + ']'


def sort(l):
    return list(sorted(l))


def digits_sort(n):
    return eval(''.join(str_sort(map(str, digits(n)))))


def str_sort(s):
    return ''.join(sort(s))


def chunk1(x, y):
    return split2(x, one_range(y))


def chunk2(x, y):
    return split2(len(x), y)


def zip_self(l):
    return zip2(l, l)


def num_zip_self(n):
    return zip_self(one_range(n))


def triplicate(x):
    return x, x, x


def assign(c, b, a):
    if not a:
        return a
    if isinstance(a, list):
        a[int(b) % len(a)] = c
        return a
    else:
        a = list(a)
        a[b % len(a)] = c
        return ''.join(map(str, a))


def swapped_assign(c, a, b):
    return assign(c, b, a)


def length_assign(c, b, a):
    return assign(c, len(b), a)


def num_assign(c, b, a):
    return assign(c, b, digits(a))


def vectorised_assign(c, b, a):
    for i in b:
        if isinstance(i, (str, list)):
            a = length_assign(c, i, a)
        else:
            a = assign(c, i, a)
    return a


def to_binary(n):
    if n < 0:
        return '-' + to_binary(abs(n))
    return bin(int(n))[2:]


def ord_bin(s):
    return [to_binary(ord(c)) for c in s]


def combinations1(n, x):
    if n > len(x):
        n = len(x)
    if n < 1:
        n = 1
    return list(map(list, itertools.combinations(x, n)))


def combinations2(x, n):
    return combinations1(n, x)


def combinations3(x, y):
    return combinations1(x, one_range(y))


def set_union(x, y):
    r = []
    for i in y:
        if i not in r:
            r.append(i)
    for j in x:
        if j not in r:
            r.append(j)
    return r


def from_list_of_digits(b, n):
    return from_list_of_digits_2(n, int(b))


def divmod2(x, y):
    return list(divmod(y, x))


def length_range(x):
    return one_range(len(x))


def exclusive_one_range(n):
    return one_range(n)[:-1]


def prime_factor_exponents(n):
    p = prime_factors(n)
    r = []
    for i in range(n + 1):
        if is_prime(i):
            r.append(p.count(i))
    x = r[::-1]
    for k, j in enumerate(x):
        if j:
            return x[k:][::-1]
    return []


def group_consecutive(l):
    last = None
    r = []
    for i in l:
        if i != last:
            r.append([i])
        else:
            r[-1].append(i)
        last = i
    if isinstance(l, str):
        return [*map(''.join, r)]
    return r


def consecutive_digits(n):
    return group_consecutive(digits(n))


def num_head_remove(x):
    return int(str(x)[1:])


def head_remove(x):
    if not x:
        return x
    return x[1:]


def lcm(l):
    l = [int(i) for i in l if isinstance(i, (int, float)) or str(i).isdigit()]
    if not l:
        return 0
    n = 1
    while 1:
        if all(n % j == 0 for j in l):
            return n
        n += 1


def digits_lcm(n):
    return lcm([int(i) for i in str(n) if i in '0123456789'])


def ords_lcm(s):
    return lcm(list(map(ord, s)))


def median(lst):
    l = []
    for i in lst:
        if isinstance(i, (int, float)):
            l.append(i)
        else:
            try:
                l.append(int(i))
            except:
                try:
                    l.append(float(i))
                except:
                    pass
    if not l:
        return []
    if len(l) % 2:
        return sorted(l)[len(l) // 2]
    s = sorted(l)[len(l) // 2 - 1: len(l) // 2 + 1]
    return mean(s)


def num_median(n):
    return median(digits(n))


def ord_median(s):
    return median(list(map(ord, s)))


def round_decimal_places(a, b):
    return round(b, int(a))


def range_intersection1(a, b):
    return [i for i in b if i in one_range(a)]


def range_intersection2(a, b):
    return [i for i in one_range(b) if i in a]


def set_intersection(a, b):
    return [i for i in b if i in a]


def permutations1(n, x):
    if n > len(x):
        n = len(x)
    if n < 1:
        n = 1
    return list(map(list, itertools.permutations(x, n)))


def permutations2(x, n):
    return permutations1(n, x)


def permutations3(x, y):
    return permutations1(x, one_range(y))


def set_difference(x, y):
    r = []
    for i in y:
        if i not in r:
            if i not in x:
                r.append(i)
    return r


def rjust1(x, n, y):
    return str(y).rjust(abs(int(n)) or 1, (str(x) + ' ')[0])


def rjust2(n, x, y):
    return str(y).rjust(abs(int(n)) or 1, (str(x) + ' ')[0])


def rjust3(x, y, z):
    return str(z).rjust(len(y) or 1, (str(x) + ' ')[0])


def prefixes(l):
    r = []
    for i in range(1, len(l) + 1):
        r.append(l[:i])
    return r


def cumsum(l):
    if any(not isinstance(i, (int, float)) for i in l):
        if not all(isinstance(j, list) for j in l):
            l = [*map(str, l)]
    if not l:
        return []
    r = []
    for i in range(1, len(l) + 1):
        r.append(it_sum(l[:i]))
    return r


def numcumsum(n):
    return cumsum(digits(n))


def num_tail_remove(x):
    return int(str(x)[:-1])


def tail_remove(x):
    if not x:
        return x
    return x[:-1]


def chunk3(n, l):
    x, y = len(l), abs(int(n))
    return list(filter(bool, [l[i:i + y] for i in range(0, (x // y + 1) * y, y)]))


def chunk4(l, n):
    return chunk3(n, l)


def chunk5(x, y):
    return chunk3(x, one_range(y))


def chunk6(x, y):
    return chunk3(len(x), y)


def length_range_no_pop(l):
    return (l, one_range(len(l)))


def num_length_range(n):
    return n, one_range(len(str(n)))


def length_range_0(l):
    return (l, lowered_range(len(l)))


def num_length_range_0(n):
    return n, lowered_range(len(str(n)))


def character_multiply1(s, n):
    return ''.join(c * n for c in s)


def character_multiply2(n, s):
    return character_multiply1(s, n)


def regex_split(x, y):
    return re.split(x, y)


def double(x):
    return x * 2


def exactly_not_equal(x, y):
    return int(y != x)


def to_hex(n):
    if n < 0:
        return '-' + to_hex(abs(n))
    return hex(int(n))[2:]


def ord_hex(s):
    return [to_hex(ord(c)) for c in s]


def reciprocal(n):
    return 1 / n


def remove_non_alphabets(s):
    return ''.join(c for c in s if c.isalpha())


def num_bifurcate(n):
    return n, digit_reverse(n)


def bifurcate(x):
    return x, reverse(x)


def fallback_lcm(a, b):
    x, y = abs(int(a)), abs(int(b))
    return (x * y) // dyadic_gcd(x, y)


def dyadic_lcm(a, b):
    try:
        return math.lcm(abs(int(a)), abs(int(b)))
    except:
        try:
            return fallback_lcm(a, b)
        except:
            return 1


def ordinal_lcm1(a, b):
    return [dyadic_lcm(ord(c), b) for c in a]


def ordinal_lcm2(a, b):
    return ordinal_lcm1(b, a)


def ordinal_lcm3(a, b):
    return [dyadic_lcm(ord(x), ord(y)) for x, y in zip(a, b)]


def num_mirror(n):
    r = str(n) + str(n)[::-1]
    try:
        return eval(r)
    except:
        return r


def mirror(x):
    return x + x[::-1]


def str_transliterate(a, b, c):
    for i, j in zip(a, b):
        c = str(c).replace(str(j), str(i))
    return str(c)


def list_transliterate(a, b, c):
    for i, j in zip(a, b):
        c = list_replace(i, j, c)
    return c


def str_transliterate_overload_1(a, b, c):
    for i, j in zip(a, str(b)):
        c = str(c).replace(str(j), str(i))
    return str(c)


def list_transliterate_overload_1(a, b, c):
    for i, j in zip(a, str(b)):
        c = list_replace(i, j, c)
    return c


def str_transliterate_overload_2(a, b, c):
    for i, j in zip(str(a), b):
        c = str(c).replace(str(j), str(i))
    return str(c)


def list_transliterate_overload_2(a, b, c):
    for i, j in zip(str(a), b):
        c = list_replace(i, j, c)
    return c


def str_transliterate_overload_3(a, b, c):
    for i, j in zip(str(a), str(b)):
        c = str(c).replace(str(j), str(i))
    return str(c)


def list_transliterate_overload_3(a, b, c):
    for i, j in zip(str(a), str(b)):
        c = list_replace(i, j, c)
    return c


def combinations_with_replacement1(n, x):
    if n > len(x):
        n = len(x)
    if n < 1:
        n = 1
    return list(map(list, itertools.combinations_with_replacement(x, n)))


def combinations_with_replacement2(x, n):
    return combinations_with_replacement1(n, x)


def combinations_with_replacement3(x, y):
    return combinations_with_replacement1(x, one_range(y))


def combinations_with_replacement4(x, y):
    return combinations_with_replacement1(len(x), y)


def square(n):
    return n ** 2


def chunk_wrap_2(s):
    return chunk3(2, s)


def cube(n):
    return n ** 3


def chunk_wrap_3(s):
    return chunk3(3, s)


def fourth(n):
    return n ** 4


def chunk_wrap_4(s):
    return chunk3(4, s)


def fifth(n):
    return n ** 5


def chunk_wrap_5(s):
    return chunk3(5, s)


def halve(n):
    if n % 2 == 0:
        return n // 2
    return n / 2


def chunk_halve(s):
    return split1(s, 2)


def deltas(lst):
    l = []
    for i in lst:
        if isinstance(i, (int, float)):
            l.append(i)
        else:
            try:
                l.append(int(i))
            except:
                try:
                    l.append(float(i))
                except:
                    pass
    if not l:
        return []
    r = []
    for x, y in zip(l, l[1:]):
        r.append(y - x)
    return r


def digit_deltas(n):
    return deltas(digits(n))


def ord_deltas(s):
    return deltas(list(map(ord, s)))


def digits_wrap(n):
    return [digits(n)]


def list_wrap(s):
    return [[*s]]


def transpose(lst):
    l = []
    for i in lst:
        if isinstance(i, (list, str)):
            l.append(list(i))
        elif isinstance(i, (int, float)):
            l.append(digits(i))
    return list(map(list, zip(*l)))


def rotate_left(n, l):
    if not l:
        return l
    rot = int(n) % len(l)
    return l[rot:] + l[:rot]


def rotate_right(n, l):
    if not l:
        return l
    rot = int(n) % len(l)
    return l[-rot:] + l[:-rot]


def swapped_rotate_left(l, n):
    return rotate_left(n, l)


def swapped_rotate_right(l, n):
    return rotate_right(n, l)


def len_rotate_left(x, y):
    return rotate_left(len(x), y)


def len_rotate_right(x, y):
    return rotate_right(len(x), y)


def num_rotate_left(n, x):
    r = rotate_left(n, str(x))
    try:
        return eval(r)
    except:
        return r


def num_rotate_right(n, x):
    r = rotate_right(n, str(x))
    try:
        return eval(r)
    except:
        return r


def rotate_left_once(l):
    return rotate_left(1, l)


def rotate_right_once(l):
    return rotate_right(1, l)


def num_rotate_left_once(l):
    return num_rotate_left(1, l)


def num_rotate_right_once(l):
    return num_rotate_right(1, l)


def left_bit_shift(x, y):
    if x < 0:
        return right_bit_shift(-x, y)
    return int(y) << int(x)


def right_bit_shift(x, y):
    if x < 0:
        return left_bit_shift(-x, y)
    return int(y) >> int(x)


def uninterleave2(n, l):
    if not n:
        return l
    return [l[i::n] if i < len(l) else l * 0 for i in range(abs(int(n)))]


def num_uninterleave2(x, y):
    return uninterleave2(x, digits(y))


def swapped_uninterleave(l, n):
    return uninterleave2(n, l)


def len_uninterleave(x, y):
    return uninterleave2(len(x), y)


def num_to_alphabet(n):
    return chr(64 + (int(n) % 26))


def alphabet_to_num(s):
    return [
        ord(c) - 64 if 'A' <= c <= 'Z' else (
            ord(c) - 96 if 'a' <= c <= 'z' else 0
        ) for c in s
    ]


def nPr(r, n):
    n, r = int(n), int(r)
    f = math.factorial
    return f(n) // f(n - r)


def string_contains(x, y):
    if isinstance(x, list):
        return [string_contains(i, y) for i in x]
    return int(str(x) in str(y))


def list_contains(x, y):
    if isinstance(x, list):
        return [list_contains(i, y) for i in x]
    return int(x in y)


def list_remove_at_index(n, l):
    if not l:
        return l
    x = l.copy()
    x.pop(int(n) % len(x))
    return x


def str_remove_at_index(n, s):
    return "".join(list_remove_at_index(n, [*str(s)]))


def swapped_list_remove_at_index(l, n):
    return list_remove_at_index(n, l)


def swapped_str_remove_at_index(s, n):
    return str_remove_at_index(n, s)


def unique_prime_factors(n):
    return uniquify_lst(prime_factors(n))


def palindromise(l):
    return l + l[:-1][::-1]


def num_palindromise(n):
    s = palindromise(str(n))
    try:
        return eval(s)
    except:
        return s


def recurse_integer_partitions(n, m=1):
    for i in range(m, n // 2 + 1):
        for part in recurse_integer_partitions(n - i, i):
            yield part + [i]
    yield [n]


def integer_partitions(n):
    n = int(n)
    l = [*recurse_integer_partitions(abs(n))]
    if n < 0:
        return [[-i for i in sl] for sl in l]
    return l


def newline_join(iterable):
    if not isinstance(iterable, list):
        return str(iterable)
    return '\n'.join(map(empty_join, iterable))


def space_join(iterable):
    if not isinstance(iterable, list):
        return str(iterable)
    return ' '.join(map(empty_join, iterable))


def prepend(x, l):
    return [x] + l


def swapped_prepend(l, x):
    return prepend(x, l)


def prepend2(x, y):
    return [x] + _digits(y)


def transpose_with_filler(f, l):
    matrix = [
        [*x] if isinstance(x, (list, str)) else digits(x)
        for x in l
    ]
    transposed = itertools.zip_longest(*matrix, fillvalue=f)
    return list(map(list, transposed))


def swapped_transpose_with_filler(l, f):
    return transpose_with_filler(f, l)


def digits_wrap_2(x, y):
    return y, digits_wrap(x)


def list_wrap_2(x, y):
    return y, list_wrap(x)


def pop_(*_):
    return ()


def boolify(x):
    return int(bool(x))


def counts(l):
    r = []
    for i in uniquify_lst(l):
        r.append([i, l.count(i)])
    return r


def num_counts(n):
    return counts(digits(n))


def parity(n):
    return n % 2


def second_half(s):
    return chunk_halve(s)[-1]


def num_prefixes(n):
    return prefixes(digits(n))


def increment_twice(n):
    return n + 2


def grade_up(l):
    return [*map(lambda x: x[0], sorted(enumerate(l[:]), key=lambda x: x[-1]))]


def head_slice(n, l):
    return l[:n]


def num_head_slice(x, y):
    s = head_slice(x, str(y))
    try:
        return eval(s)
    except:
        return s


def swapped_head_slice(l, n):
    return head_slice(n, l)


def len_head_slice(x, y):
    return head_slice(len(x), y)


def tail_slice(n, l):
    return l[-n:]


def num_tail_slice(x, y):
    s = tail_slice(x, str(y))
    try:
        return eval(s)
    except:
        return s


def swapped_tail_slice(l, n):
    return tail_slice(n, l)


def len_tail_slice(x, y):
    return tail_slice(len(x), y)


def powerset(l):
    r = [l * 0]
    p = [[]]
    for i in l:
        n = [x + [i] for x in p]
        p += [s[:] for s in n]
        for j in n:
            r.append(j if not isinstance(l, str) else "".join(j))
    return r


def num_powerset(n):
    return powerset(one_range(n))


def randint(n):
    if n == 0:
        return 0
    return random.choice(one_range(n))


def randchoice(l):
    if not l:
        return l
    return random.choice(l)


def square_root(i):
    n = abs(i)
    x, y = math.sqrt(n), math.isqrt(n)
    if x == y:
        return y
    return x


def every_second_item(s):
    return s[::2]


def sign(n):
    if n < 0:
        return -1
    elif n > 0:
        return 1
    else:
        return 0


def sentence_case(s):
    r = ''
    b = True
    for c in s:
        if b:
            r += c.upper()
        else:
            r += c.lower()
        if c in '?!.':
            b = True
        elif c == ' ' and b == True:
            pass
        else:
            b = False
    return r


def sum_each(l):
    r = []
    for i in l:
        if isinstance(i, (list, str)):
            r.append(it_sum(i))
        else:
            r.append(digit_sum(i))
    return r


def all_equal(l):
    return int(len({*l}) <= 1)


def num_all_equal(n):
    return all_equal(str(n))


def symmetric_set_difference(x, y):
    r = uniquify_lst([i for i in y if i not in x] + [j for j in x if j not in y])
    if (type(x), type(y)) == (str, str):
        return ''.join(r)
    return r


def range_ssd1(x, y):
    return symmetric_set_difference(digits(x), digits(y))


def range_ssd2(n, x):
    if isinstance(x, str):
        return symmetric_set_difference(str(n), x)
    return symmetric_set_difference(digits(n), x)


def range_ssd3(x, n):
    if isinstance(x, str):
        return symmetric_set_difference(x, str(n))
    return symmetric_set_difference(x, digits(n))


def equals_one(x):
    return int(x == 1)


def enclose(l):
    return l + l[:1]


def num_enclose(n):
    s = enclose(str(n))
    try:
        return eval(s)
    except:
        return s


def add_digits1(l, x):
    return _digits(x) + l


def add_digits2(x, l):
    return l + _digits(x)


def add_digits3(x, y):
    return _digits(y) + _digits(x)


def corresponding_filter(x, y):
    r = []
    for i, j in zip(y, x):
        if j:
            r.append(i)
    if isinstance(y, str):
        return ''.join(r)
    return r


def corresponding_digit_filter_1(x, y):
    return corresponding_filter(digits(x), digits(y))


def corresponding_digit_filter_2(n, l):
    return corresponding_filter(digits(n), l)


def corresponding_digit_filter_3(l, n):
    return corresponding_filter(l, digits(n))


def list_sort_uniquify(l):
    return sort(uniquify_lst(l))


def str_sort_uniquify(s):
    return str_sort(uniquify_str(s))


def num_sort_uniquify(n):
    return list_sort_uniquify(digits(n))


def to_roman_numerals(num):
    ret = ''
    if num < 0:
        num = -num
        ret += '-'
    symbols = {'M': 1000, 'CM': 900, 'D': 500, 'CD': 400, 'C': 100, 'XC': 90, 'L': 50, 'XL': 40, 'X': 10, 'IX': 9,
               'V': 5, 'IV': 4, 'I': 1}
    for s, n in symbols.items():
        while num >= n:
            ret += s
            num -= n
    return ret


def from_roman_numerals(s):
    ret = 0
    i = 0
    while i < len(s):
        c = s[i]
        if c == 'M':
            ret += 1000
        elif c == 'D':
            ret += 500
        elif c == 'L':
            ret += 50
        elif c == 'V':
            ret += 5
        elif c == 'C':
            i += 1
            if s[i] == 'M':
                ret += 900
            else:
                ret += 100
                i -= 1
        elif c == 'X':
            i += 1
            if s[i] == 'C':
                ret += 90
            else:
                ret += 10
                i -= 1
        elif c == 'I':
            i += 1
            if s[i] == 'X':
                ret += 9
            else:
                ret += 1
                i -= 1
        i += 1
    return ret


def length_of_each(l):
    r = []
    for i in l:
        if isinstance(i, (int, float)):
            r.append(len(str(i)))
        else:
            r.append(len(i))
    return r


def complement(n):
    return 1 - n


def is_vowel(s):
    return [int(c.lower() in 'aeiou') for c in s]


def islower(s):
    if not s:
        return 1
    return int(s.islower())


def cmp(x, y):
    return sign(y - x)


def zfill1(n, s):
    if n < 0:
        return zfill1(abs(n), s[::-1])[::-1]
    return s.zfill(n)


def zfill2(s, n):
    return zfill1(n, s)


def surround(x, y):
    return x + y + x


def digits_repeat1(n, x):
    return _digits(x) * int(n)


def digits_repeat2(x, n):
    return digits_repeat1(n, x)


def length_repeat1(x, y):
    if isinstance(y, (int, float)):
        y = str(y)
    return x * len(y)


def length_repeat2(x, y):
    if isinstance(x, (int, float)):
        y = str(x)
    return _digits(y) * len(x)


def shuffle(l):
    return random.sample(l, len(l))


def str_shuffle(s):
    return ''.join(shuffle(s))


def range_shuffle(n):
    return shuffle(one_range(n))


def spaces(n):
    return int(n) * ' '


def suffixes(l):
    r = []
    for i in range(1, len(l) + 1):
        r.append(l[-i:])
    return r


def triple_swap(a, b, c):
    return a, c, b


def cartesian_power(n, l):
    return list(map(list, itertools.product(*[l] * abs(int(n)))))


def swapped_cartesian_power(l, n):
    return cartesian_power(n, l)


def range_cartesian_power(x, y):
    return cartesian_power(x, one_range(y))


def head_remove_2(n, l):
    return l[n:]


def num_head_remove_2(x, y):
    s = head_remove_2(x, str(y))
    try:
        return eval(s)
    except:
        return s


def swapped_head_remove_2(l, n):
    return head_remove_2(n, l)


def len_head_remove_2(x, y):
    return head_remove_2(len(x), y)


def tail_remove_2(n, l):
    return l[:-n]


def num_tail_remove_2(x, y):
    s = tail_remove_2(x, str(y))
    try:
        return eval(s)
    except:
        return s


def swapped_tail_remove_2(l, n):
    return tail_remove_2(n, l)


def len_tail_remove_2(x, y):
    return tail_remove_2(len(x), y)


def reverse_each(l):
    r = []
    for i in l:
        if isinstance(i, (str, list)):
            r.append(i[::-1])
        else:
            s = str(i)[::-1]
            try:
                r.append(eval(s))
            except:
                r.append(s)
    return r


def listify(x):
    if isinstance(x, (list, str)):
        return list(copy.deepcopy(x))
    return one_range(x)
